import pytest
from unittest.mock import MagicMock, patch
from unittest import mock
import os
import tempfile
from urllib.parse import urljoin

from geobox.enums import LayerType
from geobox.raster import Raster
from geobox.user import User
from geobox.task import Task
from geobox.vectorlayer import VectorLayer


def test_init(api, mock_raster_data):
    """Test Raster initialization."""
    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
    assert raster.uuid == mock_raster_data['uuid']
    assert raster.name == mock_raster_data['name']
    assert raster.data == mock_raster_data
    assert raster.endpoint == f"{Raster.BASE_ENDPOINT}{mock_raster_data['uuid']}/"


def test_repr(api, mock_raster_data):
    """Test Raster string representation."""
    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
    assert repr(raster) == f"Raster(uuid={raster.uuid}, name={raster.name})"


def test_getattr(api, mock_raster_data):
    """Test getting attributes of Raster object."""
    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
    assert raster.name == mock_raster_data['name']
    assert raster.display_name == mock_raster_data['display_name']
    assert raster.description == mock_raster_data['description']


def test_get_rasters(api, mock_raster_data):
    """Test getting list of rasters."""
    mock_response = [
        mock_raster_data,
        mock_raster_data
    ]
    api.get.return_value = mock_response
    
    rasters = Raster.get_rasters(api)
    assert len(rasters) == 2
    assert rasters[0].name == mock_raster_data['name']
    assert rasters[1].name == mock_raster_data['name']
    assert rasters[0].uuid == mock_raster_data['uuid']
    assert rasters[1].uuid == mock_raster_data['uuid']
    assert rasters[1].data == mock_raster_data
    api.get.assert_called_once_with(f'{Raster.BASE_ENDPOINT}?f=json&return_count=False&skip=0&limit=100&shared=False')


def test_get_rasters_by_ids(api, mock_raster_data):
    mock_response = [{**mock_raster_data, **{'id': 1}}, {**mock_raster_data, **{'id': 2}}]
    api.get.return_value = mock_response
    
    mosaics = Raster.get_rasters_by_ids(api, [1, 2])
    
    api.get.assert_called_once_with(f'{Raster.BASE_ENDPOINT}get-rasters/?ids=%5B1%2C+2%5D')
    assert len(mosaics) == 2
    assert all(isinstance(m, Raster) for m in mosaics)
    assert mosaics[0].data == {**mock_raster_data, **{'id': 1}}
    assert mosaics[1].data == {**mock_raster_data, **{'id': 2}}


def test_get_raster(api, mock_raster_data):
    """Test getting a single raster."""
    api.get.return_value = mock_raster_data
    
    raster = Raster.get_raster(api, mock_raster_data['uuid'])
    assert raster.uuid == mock_raster_data['uuid']
    assert raster.name == mock_raster_data['name']
    assert raster.data == mock_raster_data
    api.get.assert_called_once_with(f'{Raster.BASE_ENDPOINT}{mock_raster_data["uuid"]}/?f=json')


def test_get_raster_by_name(api, mock_raster_data):
    """Test getting a single raster."""
    api.get.return_value = [mock_raster_data]
    
    raster = Raster.get_raster_by_name(api, name=mock_raster_data['name'])
    assert raster.uuid == mock_raster_data['uuid']
    assert raster.name == mock_raster_data['name']
    assert raster.data == mock_raster_data
    api.get.assert_called_once_with(f'{Raster.BASE_ENDPOINT}?f=json&q=name+%3D+%27sample_sample1.tif%27&return_count=False&skip=0&limit=100&shared=False')
    # not found
    raster = Raster.get_raster_by_name(api, name='not_found')
    assert raster == None


def test_update(api, mock_raster_data):
    """Test updating raster properties."""
    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
    api.put.return_value = {**mock_raster_data, **{'name': 'new_name', 'display_name': 'New Display Name', 'description': 'New Description'}}
    
    result = raster.update(name='new_name', display_name='New Display Name', description='New Description')
    api.put.assert_called_once_with(
        raster.endpoint,
        {'name': 'new_name', 'display_name': 'New Display Name', 'description': 'New Description'}
    )
    assert raster.name == result['name']
    assert raster.display_name == result['display_name']
    assert raster.description == result['description']


def test_delete(api, mock_raster_data):
    """Test deleting a raster."""
    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)

    endpoint = raster.endpoint
    raster.delete()
    api.delete.assert_called_once_with(endpoint)
    assert raster.uuid is None
    assert raster.endpoint is None


def test_thumbnail(api, mock_raster_data):
    """Test getting raster thumbnail URL."""
    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
    thumbnail_url = raster.thumbnail
    assert thumbnail_url == f"{api.base_url}{raster.endpoint}thumbnail"


def test_raster_info(api, mock_raster_data):
    """Test getting raster info."""
    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
    api.get.return_value = mock_raster_data
    
    info = raster.info
    assert info == mock_raster_data
    api.get.assert_called_once_with(f'{raster.endpoint}info/')


def test_get_statistics(api, mock_raster_data):
    """Test getting raster statistics."""
    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
    api.get.return_value = {'band1': {'min': 0, 'max': 255}}
    
    stats = raster.get_statistics(indexes='1')
    assert stats == {'band1': {'min': 0, 'max': 255}}
    api.get.assert_called_once_with(f'{raster.endpoint}statistics/?indexes=1')


def test_get_point(api, mock_raster_data):
    """Test getting point data from raster."""
    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
    api.get.return_value = {'value': 100}
    
    point = raster.get_point(lat=35, lng=51)
    assert point == {'value': 100}
    api.get.assert_called_once_with(f'{raster.endpoint}point?lat=35&lng=51')

    # errors
    with pytest.raises(ValueError, match='lat must be between -90 and 90'):
        raster.get_point(lat=94, lng=51)

    with pytest.raises(ValueError, match='lng must be between -180 and 180'):
        raster.get_point(lat=14, lng=251)


def test_download(api, mock_raster_data):
    """Test downloading a raster."""
    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
    mock_response = MagicMock()
    mock_response.__enter__.return_value.iter_content.return_value = [b'test content']
    mock_response.__exit__.return_value = None
    mock_response.__enter__.return_value.headers = {
        'Content-Type': 'image/tiff',
        'Content-Disposition': 'attachment; filename="test_raster.tif"'
    }
    api.get.return_value = mock_response

    with tempfile.TemporaryDirectory() as temp_dir:
        save_path = os.path.join(temp_dir, '') + '/'
        downloaded_path = raster.download(save_path=save_path)
        assert downloaded_path.endswith('.tif')
        assert os.path.exists(downloaded_path)

    mock_raster_data['name'] = None
    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
    mock_response = MagicMock()
    mock_response.__enter__.return_value.iter_content.return_value = [b'test content']
    mock_response.__exit__.return_value = None
    mock_response.__enter__.return_value.headers = {
        'Content-Type': 'image/tiff',
        'Content-Disposition': 'attachment; filename="test_raster.tif"'
    }
    api.get.return_value = mock_response

    with tempfile.TemporaryDirectory() as temp_dir:
        downloaded_path = raster.download()
        assert downloaded_path.endswith('.tif')
        assert os.path.exists(downloaded_path)


    if os.path.exists(downloaded_path):
        os.remove(downloaded_path)

    # error
    with pytest.raises(ValueError):
        raster.download(save_path='path')

    raster.uuid = None
    with pytest.raises(ValueError):
        raster.download()

    # import warning log
    import builtins
    original_import = builtins.__import__

    def mocked_import(name, globals=None, locals=None, fromlist=(), level=0):
        if name == "tqdm.auto":
            raise ImportError("No module named 'tqdm'")
        return original_import(name, globals, locals, fromlist, level)

    with mock.patch("builtins.__import__", side_effect=mocked_import):
        with mock.patch("geobox.api.logger") as mock_logger:
            result = raster._create_progress_bar()
            assert result is None
            mock_logger.warning.assert_called_once()


def test_get_content(api, mock_raster_data):
    """Test getting raster content."""
    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
    mock_response = MagicMock()
    mock_response.__enter__.return_value.iter_content.return_value = [b'test content']
    mock_response.__exit__.return_value = None
    mock_response.__enter__.return_value.headers = {
        'Content-Type': 'image/tiff'
    }
    api.get.return_value = mock_response

    with tempfile.TemporaryDirectory() as temp_dir:
        save_path = os.path.join(temp_dir, '') + '/'
        downloaded_path = raster.get_content_file(save_path=save_path)
        assert downloaded_path.endswith('.tiff')
        assert os.path.exists(downloaded_path)

    # error
    raster.uuid = None
    with pytest.raises(ValueError):
        raster.get_content_file()


def test_get_render_png_url(api, mock_raster_data):
    """Test getting tile PNG URL."""
    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
    url = raster.get_render_png_url(x=10, y=20, z=1, indexes=1)
    assert url == f"{raster.api.base_url}{raster.endpoint}render/1/10/20.png?indexes=1"
    # apikey
    api.access_token = ''
    api.apikey = 'apikey_1234'
    url = raster.get_render_png_url(x=10, y=20, z=1, indexes=1)
    assert url == f"{raster.api.base_url}{raster.endpoint}render/1/10/20.png?indexes=1&apikey=apikey_1234"
    

def test_get_tile_png_url(api, mock_raster_data):
    """Test getting tile PNG URL."""
    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
    url = raster.get_tile_png_url(x=10, y=20, z=1)
    assert url == f"{raster.api.base_url}{raster.endpoint}tiles/1/10/20.png"
    # apikey
    api.access_token = ''
    api.apikey = 'apikey_1234'
    url = raster.get_tile_png_url(x=10, y=20, z=1)
    assert url == f"{raster.api.base_url}{raster.endpoint}tiles/1/10/20.png?apikey=apikey_1234"


def test_get_tile_pbf_url(api, mock_raster_data):
    """Test getting tile URL."""
    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
    url = raster.get_tile_pbf_url(x=10, y=20, z=1)
    assert url == f"{raster.api.base_url}{raster.endpoint}tiles/1/10/20.pbf"
    # apikey
    api.access_token = ''
    api.apikey = 'apikey_1234'
    url = raster.get_tile_pbf_url(x=10, y=20, z=1)
    assert url == f"{raster.api.base_url}{raster.endpoint}tiles/1/10/20.pbf?apikey=apikey_1234"


def test_get_tile_json(api, mock_raster_data):
    """Test getting raster tile JSON."""
    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
    api.get.return_value = mock_raster_data
    response = raster.get_tile_json()
    assert response == mock_raster_data
    api.get.assert_called_once_with(f'{raster.endpoint}tilejson.json')


def test_wmts(api, mock_raster_data):
    """Test getting WMTS URL."""
    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
    url = raster.wmts(scale=1)
    assert url == f'{raster.api.base_url}{raster.endpoint}wmts/?scale=1'
    # apikey
    api.access_token = ''
    api.apikey = 'apikey_1234'
    url = raster.wmts(scale=1)
    assert url == f'{raster.api.base_url}{raster.endpoint}wmts/?scale=1&apikey=apikey_1234'


def test_settings(api, mock_raster_data):
    """Test getting raster settings."""
    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
    api.get.return_value = mock_raster_data
    settings = raster.settings
    assert settings == mock_raster_data
    api.get.assert_called_once_with(f'{raster.endpoint}settings/?f=json')


def test_update_settings(api, mock_raster_data):
    raster1 = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
    raster2 = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
    settings = raster2.settings
    api.put.return_value = settings
    raster1.update_settings(raster2.settings)
    api.put.assert_called_once_with(f'{raster1.endpoint}settings/', settings)


def test_set_settings(api, mock_raster_data):
    """Test setting raster settings."""
    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
    api.get.return_value = {
        "visual_settings": {
            "nodata": None,
            "indexes": [
            1,
            2,
            3
            ],
            "rescale": [
            [
                0,
                9848
            ],
            [
                0,
                9636
            ],
            [
                0,
                9409
            ]
            ],
            "colormap_name": None,
            "colormap": None,
            "color_formula": None,
            "expression": None,
            "exaggeration": 10
        },
        "tile_settings": {
            "min_zoom": 0,
            "max_zoom": 13,
            "use_cache": True,
            "cache_until_zoom": 17
        }
        }
    raster.set_settings(nodata=0, indexes=[1], rescale=[[0, 255]], colormap_name='viridis', color_formula='R*255', expression='R', exaggeraion=1, min_zoom=0, max_zoom=18, use_cache=True, cache_until_zoom=10)
    expected_data = {'visual_settings': 
                        {'nodata': 0, 
                            'indexes': [1], 
                            'rescale': [[0, 255]], 
                            'colormap_name': 'viridis', 
                            'colormap': None, 
                            'color_formula': 'R*255', 
                            'expression': 'R', 
                            'exaggeration': 10, 
                            'exaggeraion': 1
                        }, 
                        'tile_settings': 
                            {'min_zoom': 0, 
                                'max_zoom': 18, 
                                'use_cache': True, 
                                'cache_until_zoom': 10
                            }
                    }
    api.put.assert_called_once_with(
        f"{raster.endpoint}settings/",
        expected_data
    )


def test_share(api, mock_raster_data, mock_user_data):
    """Test file sharing."""
    with patch.object(api, 'post'):
        raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
        users = [
            User(api, user_id=mock_user_data['1']['id'], data=mock_user_data['1']),
            User(api, user_id=mock_user_data['2']['id'], data=mock_user_data['2'])
        ]
        raster.share(users=users)
        api.post.assert_called_once_with(
            f'{raster.endpoint}share/',
            {'user_ids': [1, 2]}, is_json=False
        )


def test_unshare(api, mock_raster_data, mock_user_data):
    """Test file unsharing."""
    with patch.object(api, 'post'):
        raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
        users = [
            User(api, user_id=mock_user_data['1']['id'], data=mock_user_data['1']),
            User(api, user_id=mock_user_data['2']['id'], data=mock_user_data['2'])
        ]
        raster.unshare(users=users)
        
        # Verify the API call
        api.post.assert_called_once_with(
            f'{raster.endpoint}unshare/',
            {'user_ids': [1, 2]}, is_json=False
        )


def test_get_shared_users(api, mock_raster_data, mock_user_data):
    """Test getting shared users."""
    mock_response = [
        mock_user_data['1'],
        mock_user_data['2']
    ]
    
    with patch.object(api, 'get', return_value=mock_response):
        raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
        result = raster.get_shared_users(search='user', limit=2)

        api.get.assert_called_once_with(
            f'{raster.endpoint}shared-with-users/?search=user&skip=0&limit=2'
        )

        assert len(result) == 2
        assert result[0].first_name == 'test 1'
        assert result[0].last_name == 'test 1'
        assert result[1].first_name == 'test 2'
        assert result[1].last_name == 'test 2'


def test_seed_cache(api, mock_raster_data, mock_success_task_data):
    """Test seeding cache."""
    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
    api.post.return_value = [{'task_id': mock_success_task_data['id']}]
    api.get_task.return_value = Task(api, mock_success_task_data['uuid'], mock_success_task_data)
    
    task = raster.seed_cache(from_zoom=0, to_zoom=10, extent=[0, 0, 100, 100], workers=2)[0]
    api.post.assert_called_once_with(f'{raster.endpoint}cache/seed/', {
        'from_zoom': 0,
        'to_zoom': 10,
        'extent': [0, 0, 100, 100],
        'workers': 2
    })
    assert isinstance(task, Task)


def test_clear_cache(api, mock_raster_data):
    """Test clearing cache."""
    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
    raster.clear_cache()
    api.post.assert_called_once_with(f"{raster.endpoint}cache/clear/")

def test_cache_size(api, mock_raster_data):
    """Test getting cache size."""
    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
    api.post.return_value = 1024
    
    size = raster.cache_size
    assert size == 1024
    api.post.assert_called_once_with(f"{raster.endpoint}cache/size/") 


def test_to_async(api, async_api, mock_raster_data):
    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
    async_instance = raster.to_async(async_api)
    assert async_instance.api == async_api 


def test_profile(api, mock_raster_data):
    raster = Raster(api, uuid=mock_raster_data['uuid'], data=mock_raster_data)
    api.post.return_value = {'key': 'value'}
    result = raster.profile(polyline=[[0, 0], [10, 10]], number_of_samples=200)
    assert result == {'key': 'value'}
    api.post.assert_called_once_with(endpoint=f'rasters/{raster.uuid}/profile/', payload={'polyline': [[0, 0], [10, 10]], 'number_of_samples': 200, 'include_distance': True, 'treat_nodata_as_null': True})


from urllib.parse import urljoin
import pytest
from unittest.mock import patch, MagicMock
from datetime import datetime

from geobox.map import Map
from geobox.model3d import Model
from geobox.user import User
from geobox.task import Task
from geobox.attachment import Attachment
from geobox.file import File


def test_init(api, mock_map_data):
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    assert map.name == mock_map_data['name']
    assert map.uuid == mock_map_data['uuid']
    assert map.endpoint == f'maps/{mock_map_data["uuid"]}/'


def test_repr(api, mock_map_data):
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    assert repr(map) == f"Map(uuid={map.uuid}, name={map.name})"


def test_attribute_access(api, mock_map_data):
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    assert map.display_name == mock_map_data['display_name']
    assert map.description == mock_map_data['description']
    assert map.extent == mock_map_data['extent']
    assert map.created_at == datetime.strptime(mock_map_data['created_at'], "%Y-%m-%dT%H:%M:%S.%f")
    assert map.last_modified_at == datetime.strptime(mock_map_data['last_modified_at'], "%Y-%m-%dT%H:%M:%S.%f")

    with pytest.raises(AttributeError):
        _ = map.nonexistent_attribute


@patch('geobox.map.urljoin')
@patch('geobox.map.urlencode')
def test_get_maps(mock_urlencode, mock_urljoin, api, mock_map_data):
    mock_urlencode.return_value = 'f=json'
    mock_urljoin.return_value = 'maps/?f=json'
    
    # Mock API response
    api.get.return_value = [mock_map_data, mock_map_data]

    maps = Map.get_maps(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    assert len(maps) == 2
    assert maps[0].uuid == mock_map_data['uuid']
    assert maps[0].data == mock_map_data
    api.get.assert_called_once_with(f'{Map.BASE_ENDPOINT}?f=json&return_count=False&skip=0&limit=10&shared=False')

    # Test with return_count
    api.reset_mock()
    api.get.return_value = 5
    count = Map.get_maps(api, return_count=True)
    assert count == 5
    api.get.assert_called_once_with(f'{Map.BASE_ENDPOINT}?f=json&return_count=True&skip=0&limit=10&shared=False')


def test_create_map(api):
    map_data = {
        'uuid': 'test-uuid-123',
        'name': 'Test Map',
        'display_name': 'Test Display Name',
        'description': 'Test Description',
        'extent': [10, 20, 30, 40],
        'thumbnail': 'https://example.com/thumbnail.png',
        'style': {'type': 'style'}
    }
    api.post.return_value = map_data

    map = Map.create_map(
        api,
        name='Test Map',
        display_name='Test Display Name',
        description='Test Description',
        extent=[10, 20, 30, 40],
        thumbnail='https://example.com/thumbnail.png',
        style={'type': 'style'}
    )

    assert map.uuid == map_data['uuid']
    assert map.data == map_data
    del map_data['uuid']
    api.post.assert_called_once_with(Map.BASE_ENDPOINT, map_data)


@patch('geobox.map.urljoin')
@patch('geobox.map.urlencode')
def test_get_map(mock_urlencode, mock_urljoin, api, mock_map_data):
    mock_urlencode.return_value = 'f=json'
    mock_urljoin.return_value = f'maps/{mock_map_data["uuid"]}/?f=json'
    api.get.return_value = mock_map_data

    map = Map.get_map(api, mock_map_data['uuid'])
    assert map.uuid == mock_map_data['uuid']
    assert map.data == mock_map_data
    api.get.assert_called_once_with(f"{map.endpoint}?f=json")


@patch('geobox.map.urljoin')
@patch('geobox.map.urlencode')
def test_get_map_by_name(mock_urlencode, mock_urljoin, api, mock_map_data):
    mock_urlencode.return_value = 'f=json'
    mock_urljoin.return_value = f'{Map.BASE_ENDPOINT}?f=json'
    api.get.return_value = [mock_map_data]

    map = Map.get_map_by_name(api, name=mock_map_data['name'])
    assert map.uuid == mock_map_data['uuid']
    assert map.data == mock_map_data
    api.get.assert_called_once_with(f"{Map.BASE_ENDPOINT}?f=json&q=name+%3D+%27my_map%27&return_count=False&skip=0&limit=10&shared=False")
    # not found
    map = Map.get_map_by_name(api, name='not_found')
    assert map == None


def test_update(api, mock_map_data):
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    updated_data = {
        'name': 'Updated Map',
        'display_name': 'Updated Display Name',
        'description': 'Updated Description'
    }
    api.put.return_value = updated_data

    response = map.update(
        name='Updated Map',
        display_name='Updated Display Name',
        description='Updated Description'
    )

    assert response == updated_data
    assert map.name == 'Updated Map'
    assert map.display_name == 'Updated Display Name'
    assert map.description == 'Updated Description'
    api.put.assert_called_once_with(map.endpoint, updated_data)


def test_delete(api, mock_map_data):
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    endpoint = map.endpoint
    map.delete()
    assert map.uuid is None
    assert map.endpoint is None
    api.delete.assert_called_once_with(endpoint)


def test_style(api, mock_map_data):
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    style_data = {'type': 'style', 'version': '1.0'}
    api.get.return_value = style_data
    style = map.style
    assert style == style_data
    api.get.assert_called_once_with(f"{map.endpoint}style/")


def test_thumbnail(api, mock_map_data):
    api.access_token = ''
    api.apikey = 'apikey_1234'
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    thumbnail_url = map.thumbnail
    assert thumbnail_url == f'{api.base_url}{map.endpoint}thumbnail.png?apikey=apikey_1234'
    # error
    api.apikey = ''
    with pytest.raises(ValueError):
        map.thumbnail


def test_set_readonly(api, mock_map_data):
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    response_data = {'readonly': True}
    api.post.return_value = response_data

    map.set_readonly(True)
    assert map.readonly is True
    api.post.assert_called_once_with(urljoin(map.endpoint, 'setReadonly/'), response_data, is_json=False)


def test_set_multiuser(api, mock_map_data):
    map = Map(api, uuid='test-uuid-123', data=mock_map_data)
    response_data = {'multiuser': True}
    api.post.return_value = response_data

    map.set_multiuser(True)
    assert map.multiuser is True
    api.post.assert_called_once_with(urljoin(map.endpoint, 'setMultiuser/'), response_data, is_json=False)


def test_wmts(api, mock_map_data):
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    wmts_url = map.wmts(scale=1)
    assert wmts_url == f'{api.base_url}{map.endpoint}wmts/?scale=1'
    # apikey
    api.access_token = ''
    api.apikey = 'apikey_1234'
    wmts_url = map.wmts(scale=1)
    assert wmts_url == f"{api.base_url}{map.endpoint}wmts/?scale=1&apikey=apikey_1234"


def test_share(api, mock_map_data, mock_user_data):
    """Test file sharing."""
    with patch.object(api, 'post'):
        map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
        users = [
            User(api, user_id=mock_user_data['1']['id'], data=mock_user_data['1']),
            User(api, user_id=mock_user_data['2']['id'], data=mock_user_data['2'])
        ]
        map.share(users=users)
        api.post.assert_called_once_with(
            f'{map.endpoint}share/',
            {'user_ids': [1, 2]}, is_json=False
        )


def test_unshare(api, mock_map_data, mock_user_data):
    """Test file unsharing."""
    with patch.object(api, 'post'):
        map = Map(api, uuid='test-uuid-123', data=mock_map_data)
        users = [
            User(api, user_id=mock_user_data['1']['id'], data=mock_user_data['1']),
            User(api, user_id=mock_user_data['2']['id'], data=mock_user_data['2'])
        ]
        map.unshare(users=users)
        api.post.assert_called_once_with(
            f'{map.endpoint}unshare/',
            {'user_ids': [1, 2]}, is_json=False
        )


def test_get_shared_users(api, mock_map_data, mock_user_data):
    """Test getting shared users."""
    mock_response = [
        mock_user_data['1'],
        mock_user_data['2']
    ]
    
    with patch.object(api, 'get', return_value=mock_response):
        map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
        result = map.get_shared_users(search='user', limit=2)

        api.get.assert_called_once_with(
            f'{map.endpoint}shared-with-users/?search=user&skip=0&limit=2'
        )

        assert len(result) == 2
        assert result[0].first_name == 'test 1'
        assert result[0].last_name == 'test 1'
        assert result[1].first_name == 'test 2'
        assert result[1].last_name == 'test 2'


def test_seed_cache(api, mock_map_data, mock_success_task_data):
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    task_data = {'task_id': mock_success_task_data['id']}
    api.post.return_value = task_data

    expected_data = {
        "from_zoom": 0,
        "workers": 4,
        "to_zoom": 10,
        "extent": [10, 20, 30, 40]
    }

    api.get_task.return_value = Task(api, mock_success_task_data['uuid'], mock_success_task_data)
    task = map.seed_cache(from_zoom=0, workers=4, to_zoom=10, extent=[10, 20, 30, 40])[0]
    assert task.id == mock_success_task_data['id']
    api.post.assert_called_once_with(f'{map.endpoint}cache/seed/', expected_data)


def test_update_cache(api, mock_map_data, mock_success_task_data):
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    task_data = {'task_id': mock_success_task_data['id'], 'status': 'pending'}
    api.post.return_value = [task_data]

    api.get_task.return_value = Task(api, mock_success_task_data['uuid'], mock_success_task_data)

    task = map.update_cache(from_zoom=0, to_zoom=10, extent=[10, 20, 30, 40])[0]
    assert task.id == mock_success_task_data['id']
    api.post.assert_called_once_with(f'{map.endpoint}cache/update/', 
                                    {'from_zoom': 0, 'to_zoom': 10, 'extent': [10, 20, 30, 40]})


def test_clear_cache(api, mock_map_data):
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    map.clear_cache()
    api.post.assert_called_once_with(f'{map.endpoint}cache/clear/')


def test_cache_size(api, mock_map_data):
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    api.post.return_value = 1024
    size = map.cache_size
    assert size == 1024
    api.post.assert_called_once_with(f'{map.endpoint}cache/size/')


def test_settings(api, mock_map_data):
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    api.get.return_value = {'settings': 'value'}
    settings = map.settings
    assert settings == mock_map_data['settings']


def test_update_settings(api, mock_map_data):
    map1 = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    map2 = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    settings = map2.settings
    api.put.return_value = settings
    map1.update_settings(map2.settings)
    api.put.assert_called_once_with(f'{map1.endpoint}settings/', settings)


def test_set_settings(api, mock_map_data):
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    settings = map.settings
    settings['general_settings']['base_map'] = 'osm-light'
    settings['edit_settings']['editable_layers'] = 'test'
    settings['snap_settings']['snap_cache'] = 1000
    settings['search_settings']['geosearch'] = True
    settings['marker_settings']['remvoe_unused_tags'] = False
    settings['controls'] = ["coordinates"]
    settings['terrain_settings']['exaggeration'] = 2
    settings['grid_settings']['grid_width'] = 2
    settings['view_settings']['zoom'] = 22
    settings['toc_settings'] = ['test']
    api.put.return_value = settings

    map.set_settings(base_map='osm-light', 
                        editable_layers='test', 
                        snap_cache=1000,
                        geosearch=True,
                        remove_unused_tags=False,
                        controls=["coordinates"],
                        exaggeration=2,
                        grid_width=2,
                        zoom=22,
                        toc_settings=['test'])
    assert map.settings == settings
    api.put.assert_called_once_with(f'{map.endpoint}settings/', settings)


def test_get_markers(api, mock_map_data):
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    markers_data = {
        'tags': {'#general': {'color': '#ff0000'}},
        'locations': [
            {
                'id': 1,
                'tag': '#general',
                'name': 'test',
                'geometry': [51.13162784422988, 35.766603814763045],
                'description': 'string'
            }
        ]
    }
    api.get.return_value = markers_data

    result = map.get_markers()
    assert result == markers_data
    api.get.assert_called_once_with(f'{map.endpoint}markers/')


def test_set_markers(api, mock_map_data):
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    markers_data = {
        'tags': {'#general': {'color': '#ff0000'}},
        'locations': [
            {
                'id': 1,
                'tag': '#general',
                'name': 'test',
                'geometry': [51.13162784422988, 35.766603814763045],
                'description': 'string'
            }
        ]
    }
    api.put.return_value = markers_data

    response = map.set_markers(markers_data)
    assert response == markers_data
    api.put.assert_called_once_with(f'{map.endpoint}markers/', markers_data)


def test_get_models(api, mock_map_data):
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    models_data = {
        'objects': [
            {
                'name': 'transmission_tower',
                'alias': None,
                'desc': None,
                'obj': '4079e850-055b-42a0-b35a-40d6d29ae21e',
                'loc': [53.1859045261684, 33.37762747390032, 0.0],
                'rotation': [0.0, 0.0, 0.0],
                'scale': 1.0,
                'min_zoom': 0,
                'max_zoom': 22
            }
        ]
    }
    api.get.return_value = models_data

    # Test with json=True
    result = map.get_models(json=True)
    assert result == models_data
    api.get.assert_called_once_with(f'{map.endpoint}models/')

    # Test without json=True
    result = map.get_models()
    assert len(result) == 1
    assert result[0].name == 'transmission_tower'
    assert result[0].data == models_data['objects'][0]
    assert isinstance(result[0], Model)


def test_set_models(api, mock_map_data):
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    
    # Create a mock Model object
    model = MagicMock()
    model.json = {
        'name': 'transmission_tower',
        'alias': None,
        'desc': None,
        'obj': '4079e850-055b-42a0-b35a-40d6d29ae21e',
        'loc': [53.1859045261684, 33.37762747390032, 0.0],
        'rotation': [0.0, 0.0, 0.0],
        'scale': 1.0,
        'min_zoom': 0,
        'max_zoom': 22
    }
    
    models_data = {
        'objects': [model.json]
    }
    api.put.return_value = models_data

    # Pass a dictionary as expected by the current implementation
    response = map.set_models(models_data)
    assert len(response) == 1
    assert response[0].name == 'transmission_tower'
    api.put.assert_called_once_with(f'{map.endpoint}models/', models_data)


def test_add_model(api, mock_map_data, mock_model_data, mock_map_model_data):
    # Initialize Map and Model objects with mock data
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    
    # Create a Model instance using the same UUID as in mock_map_model_data
    model = Model(
        api, 
        uuid=mock_map_model_data['obj'],  # Use the UUID from mock_map_model_data
        data={
            **mock_model_data,
            'uuid': mock_map_model_data['obj'],  # Override with expected UUID
            'name': mock_map_model_data['name']  # Override with expected name
        }
    )
    
    # Test Case 1: Adding first model to empty map
    api.get.return_value = {'objects': []}  # Empty map
    api.put.return_value = {'objects': [mock_map_model_data]}
    
    result = map.add_model(
        model,
        location=mock_map_model_data['loc'],
        rotation=mock_map_model_data['rotation'],
        scale=mock_map_model_data['scale'],
        min_zoom=mock_map_model_data['min_zoom'],
        max_zoom=mock_map_model_data['max_zoom']
    )
    
    # Verify API calls
    api.get.assert_called_once()  # Should check existing models
    api.put.assert_called_once_with(
        urljoin(map.endpoint, 'models/'),
        {'objects': [mock_map_model_data]}  # Now this will match exactly
    )
    
    # Verify return value
    assert len(result) == 1
    assert isinstance(result[0], Model)
    assert result[0].uuid == mock_map_model_data['obj']


def test_add_model_with_existing_models(api, mock_map_data, mock_model_data, mock_map_model_data):
    # Initialize Map with mock data
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    
    # Create a Model instance using the same UUID as in mock_map_model_data
    model = Model(
        api,
        uuid=mock_map_model_data['obj'],
        data={
            **mock_model_data,
            'uuid': mock_map_model_data['obj'],
            'name': mock_map_model_data['name']
        }
    )

    # Test Case: Adding a model to a map that already has models
    existing_model = {
        'name': 'existing_model',
        'alias': 'old_alias',
        'desc': 'old_description',
        'obj': 'existing-uuid-123',
        'loc': [10.0, 20.0, 0.0],
        'rotation': [0.0, 0.0, 0.0],
        'scale': 1.0,
        'min_zoom': 0,
        'max_zoom': 22
    }

    # Mock the initial GET response (map with existing models)
    api.get.return_value = {'objects': [existing_model]}
    
    # Expected response after adding new model
    expected_response = {
        'objects': [existing_model, mock_map_model_data]
    }
    api.put.return_value = expected_response

    # Call add_model
    result = map.add_model(
        model,
        location=mock_map_model_data['loc'],
        rotation=mock_map_model_data['rotation'],
        scale=mock_map_model_data['scale'],
        min_zoom=mock_map_model_data['min_zoom'],
        max_zoom=mock_map_model_data['max_zoom']
    )

    # Verify the API was called with both existing and new model
    expected_data = {
        'objects': [
            existing_model,
            {
                'name': mock_map_model_data['name'],
                'alias': None,
                'desc': None,
                'obj': mock_map_model_data['obj'],
                'loc': mock_map_model_data['loc'],
                'rotation': mock_map_model_data['rotation'],
                'scale': mock_map_model_data['scale'],
                'min_zoom': mock_map_model_data['min_zoom'],
                'max_zoom': mock_map_model_data['max_zoom']
            }
        ]
    }
    api.put.assert_called_once_with(
        urljoin(map.endpoint, 'models/'),
        expected_data
    )

    # Verify return value contains both models
    assert len(result) == 2
    assert all(isinstance(m, Model) for m in result)
    assert result[0].uuid == existing_model['obj']
    assert result[1].uuid == mock_map_model_data['obj']


def test_image_tile_url(api, mock_map_data):
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)

    assert map.image_tile_url() == f"{map.api.base_url}{map.endpoint}tiles/"+"{z}/{x}/{y}.png"
    assert map.image_tile_url(x=1, y=2, z=3, format='.pbf') == f"{map.api.base_url}{map.endpoint}tiles/3/1/2.pbf"

    # apikey
    api.access_token = ''
    api.apikey = 'apikey@1234'

    assert map.image_tile_url() == f"{map.api.base_url}{map.endpoint}tiles/"+"{z}/{x}/{y}.png?apikey=apikey@1234"


def test_export_map_to_image(api, mock_map_data, mock_success_task_data):
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)

    api.post.return_value = {
        'task_id': mock_success_task_data['uuid']
    }
    api.get_task.return_value = Task(api, mock_success_task_data['uuid'], mock_success_task_data)

    task = map.export_map_to_image(bbox=[50.275, 35.1195, 51.4459, 36.0416], width=100, height=100)
    
    api.post.assert_called_once_with('maps/da9a505b-7b51-4042-a3ba-c821e4f2cd34/export/?uuid=da9a505b-7b51-4042-a3ba-c821e4f2cd34&bbox=%5B50.275%2C+35.1195%2C+51.4459%2C+36.0416%5D&width=100&height=100')
    assert type(task) == Task


def test_get_attachments(api, mock_map_data, mock_attachment_data):
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)

    api.get.return_value = [mock_attachment_data, mock_attachment_data, mock_attachment_data]

    attachments = map.get_attachments()
    assert len(attachments) == 3
    assert type(attachments[0]) == Attachment


def test_create_attachment(api, mock_map_data, mock_attachment_data, mock_file_data):
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    file = File(api, uuid=mock_file_data['uuid'], data=mock_file_data)

    api.post.return_value = mock_attachment_data
    attachment = map.create_attachment(name='test', loc_x=10, loc_y=10, file=file)
    assert type(attachment) == Attachment
    assert attachment.data == mock_attachment_data


def test_to_async(api, async_api, mock_map_data):
    map = Map(api, uuid=mock_map_data['uuid'], data=mock_map_data)
    async_instance = map.to_async(async_api)
    assert async_instance.api == async_api 


from typing import TYPE_CHECKING, Dict, List, Optional, Union

from .base import AsyncBase

if TYPE_CHECKING:
    from .api import AsyncGeoboxClient
    from .user import AsyncUser
    from ..api import GeoboxClient
    from ..version import VectorLayerVersion

class AsyncVectorLayerVersion(AsyncBase):

    BASE_ENDPOINT = 'vectorLayerVersions/'

    def __init__(self, 
        api: 'AsyncGeoboxClient',
        uuid: str,
        data: Optional[Dict] = {}):
        """
        Initialize a vector layer version instance.

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            uuid (str): The unique identifier for the version.
            data (Dict): The data of the version.
        """
        super().__init__(api, uuid=uuid, data=data)
    

    @classmethod
    async def get_versions(cls, api: 'AsyncGeoboxClient', **kwargs) -> Union[List['AsyncVectorLayerVersion'], int]:
        """
        [async] Get list of versions with optional filtering and pagination.

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.

        Keyword Args:
            layer_id (str): the id of the vector layer.
            q (str): query filter based on OGC CQL standard. e.g. "field1 LIKE '%GIS%' AND created_at > '2021-01-01'"
            search (str): search term for keyword-based searching among search_fields or all textual fields if search_fields does not have value. NOTE: if q param is defined this param will be ignored.
            search_fields (str): comma separated list of fields for searching.
            order_by (str): comma separated list of fields for sorting results [field1 A|D, field2 A|D, …]. e.g. name A, type D. NOTE: "A" denotes ascending order and "D" denotes descending order.
            return_count (bool): Whether to return total count. default is False.
            skip (int): Number of items to skip. default is 0.
            limit (int): Number of items to return. default is 10.
            user_id (int): Specific user. privileges required.
            shared (bool): Whether to return shared versions. default is False.

        Returns:
            List[AsyncVectorLayerVersion] | int: A list of vector layer version instances or the total number of versions.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.version import AsyncVectorLayerVersion
            >>> async with AsyncGeoboxClient() as client:
            >>>     versions = await AsyncVectorLayerVersion.get_versions(client, q="name LIKE '%My version%'")
            or  
            >>>     versions = await client.get_versions(q="name LIKE '%My version%'")
        """
        params = {
            'layer_id': kwargs.get('layer_id'),
            'f': 'json',
            'q': kwargs.get('q'),
            'search': kwargs.get('search'),
            'search_fields': kwargs.get('search_fields'),
            'order_by': kwargs.get('order_by'),
            'return_count': kwargs.get('return_count', False),
            'skip': kwargs.get('skip', 0),
            'limit': kwargs.get('limit', 10),
            'user_id': kwargs.get('user_id'),
            'shared': kwargs.get('shared', False)
        }
        return await super()._get_list(api, cls.BASE_ENDPOINT, params, factory_func=lambda api, item: AsyncVectorLayerVersion(api, item['uuid'], item))


    @classmethod
    async def get_version(cls, api: 'AsyncGeoboxClient', uuid: str, user_id: int = None) -> 'AsyncVectorLayerVersion':
        """
        [async] Get a version by its UUID.

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            uuid (str): The UUID of the version to get.
            user_id (int, optional): Specific user. privileges required.

        Returns:
            AsyncVectorLayerVersion: The vector layer version object.

        Raises:
            NotFoundError: If the version with the specified UUID is not found.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.version import AsyncVectorLayerVersion
            >>> async with AsyncGeoboxClient() as client:
            >>>     version = await AsyncVectorLayerVersion.get_version(client, uuid="12345678-1234-5678-1234-567812345678")
            or  
            >>>     version = await client.get_version(uuid="12345678-1234-5678-1234-567812345678")
        """
        params = {
            'f': 'json',
            'user_id': user_id,
        }
        return await super()._get_detail(api, cls.BASE_ENDPOINT, uuid, params, factory_func=lambda api, item: AsyncVectorLayerVersion(api, item['uuid'], item))
    

    @classmethod
    async def get_version_by_name(cls, api: 'AsyncGeoboxClient', name: str, user_id: int = None) -> 'AsyncVectorLayerVersion':
        """
        [async] Get a version by name

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            name (str): the name of the version to get
            user_id (int, optional): specific user. privileges required.

        Returns:
            AsyncVectorLayerVersion | None: returns the version if a version matches the given name, else None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.version import AsyncVectorLayerVersion
            >>> async with AsyncGeoboxClient() as client:
            >>>     version = await VectorLayerView.get_version_by_name(client, name='test')
            or  
            >>>     version = await client.get_version_by_name(name='test')
        """
        versions = await cls.get_versions(api, q=f"name = '{name}'", user_id=user_id)
        if versions and versions[0].name == name:
            return versions[0]
        else:
            return None


    async def update(self, **kwargs) -> Dict:
        """
        [async] Update the version.

        Args:
            name (str): The name of the version.
            display_name (str): The display name of the version.
            description (str): The description of the version.

        Returns:
            Dict: The updated version data.

        Raises:
            ValidationError: If the version data is invalid.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.version import AsyncVectorLayerVersion
            >>> async with AsyncGeoboxClient() as client:
            >>>     version = await AsyncVectorLayerVersion.get_version(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await version.update_version(display_name="New Display Name")
        """
        data = {
            "name": kwargs.get('name'),
            "display_name": kwargs.get('display_name'),   
            "description": kwargs.get('description'),
        }
        return await super()._update(self.endpoint, data)


    async def delete(self) -> None:
        """
        [async] Delete the version.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.version import AsyncVectorLayerVersion
            >>> async with AsyncGeoboxClient() as client:
            >>>     version = await AsyncVectorLayerVersion.get_version(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await version.delete()
        """
        await super()._delete(self.endpoint)


    async def share(self, users: List['AsyncUser']) -> None:
        """
        [async] Shares the version with specified users.

        Args:
            users (List[User]): The list of user objects to share the version with.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.version import AsyncVectorLayerVersion
            >>> async with AsyncGeoboxClient() as client:
            >>>     version = await AsyncVectorLayerVersion.get_version(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     users = await client.search_users(search='John')
            >>>     await version.share(users=users)
        """
        await super()._share(self.endpoint, users)
    

    async def unshare(self, users: List['AsyncUser']) -> None:
        """
        [async] Unshares the version with specified users.

        Args:
            users (List[User]): The list of user objects to unshare the version with.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.version import AsyncVectorLayerVersion
            >>> async with AsyncGeoboxClient() as client:
            >>>     version = await AsyncVectorLayerVersion.get_version(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     users = await client.search_users(search='John')
            >>>     await version.unshare(users=users)
        """
        await super()._unshare(self.endpoint, users)


    async def get_shared_users(self, search: str = None, skip: int = 0, limit: int = 10) -> List['AsyncUser']:
        """
        [async] Retrieves the list of users the version is shared with.

        Args:
            search (str, optional): The search query.
            skip (int, optional): The number of users to skip.
            limit (int, optional): The maximum number of users to retrieve.

        Returns:
            List[AsyncUser]: The list of shared users.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.version import AsyncVectorLayerVersion
            >>> async with AsyncGeoboxClient() as client:
            >>>     version = await AsyncVectorLayerVersion.get_version(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await version.get_shared_users(search='John', skip=0, limit=10)
        """
        params = {
            'search': search,
            'skip': skip,
            'limit': limit
        }
        return await super()._get_shared_users(self.endpoint, params)

    
    def to_sync(self, sync_client: 'GeoboxClient') -> 'VectorLayerVersion':
        """
        Switch to sync version of the version instance to have access to the sync methods

        Args:
            sync_client (GeoboxClient): The sync version of the GeoboxClient instance for making requests.

        Returns:
            VectorLayerVersion: the sync instance of the version.

        Example:
            >>> from geobox import Geoboxclient
            >>> from geobox.aio import AsyncGeoboxClient
            >>> client = GeoboxClient()
            >>> async with AsyncGeoboxClient() as async_client:
            >>>     version = await async_client.get_version(async_client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     sync_version = version.to_sync(client)
        """
        from ..version import VectorLayerVersion

        return VectorLayerVersion(api=sync_client, uuid=self.uuid, data=self.data)


import pytest
from unittest.mock import patch, MagicMock
import os
from geobox.file import File

from geobox.model3d import Model
from geobox.user import User
from geobox.utils import get_save_path


def test_init(api, mock_model_data):
    model = Model(api, uuid=mock_model_data['uuid'], data=mock_model_data)
    assert model.uuid == mock_model_data['uuid']
    assert model.data == mock_model_data
    assert model.endpoint == f'3dmodels/{mock_model_data["uuid"]}/'


def test_model_repr(api, mock_model_data):
    model = Model(api, uuid=mock_model_data['uuid'], data=mock_model_data)
    assert repr(model) == f"Model(uuid={model.uuid}, name={model.name})"


def test_model_getattr(api, mock_model_data):
    model = Model(api, uuid=mock_model_data['uuid'], data=mock_model_data)
    assert model.name == mock_model_data['name']
    assert model.description == mock_model_data['description']
    
    with pytest.raises(AttributeError):
        _ = model.nonexistent_attribute


@patch('geobox.model3d.urljoin')
@patch('geobox.model3d.urlencode')
def test_get_models(mock_urlencode, mock_urljoin, api, mock_model_data):
    mock_response = [
        mock_model_data,
        mock_model_data
    ]
    api.get.return_value = mock_response
    mock_urlencode.return_value = 'f=json'
    mock_urljoin.return_value = 'https://api.geobox.ir/v1/3dmodels/?f=json'

    models = Model.get_models(api)
    assert len(models) == 2
    assert models[0].name == mock_model_data['name']
    assert models[0].data == mock_model_data
    assert models[1].name == mock_model_data['name']

    api.get.assert_called_once_with(f"{Model.BASE_ENDPOINT}?f=json&return_count=False&skip=0&limit=10&shared=False")

    # Test getting models with custom parameters
    api.reset_mock()
    api.get.return_value = {'count': 2}
    count = Model.get_models(
        api,
        q="name like 'test'",
        search="test",
        search_fields="name,description",
        order_by="name A",
        return_count=True,
        skip=0,
        limit=10,
        user_id=1,
        shared=False
    )
    assert count == 2
    api.get.assert_called_once_with(f"{Model.BASE_ENDPOINT}?f=json&q=name+like+%27test%27&search=test&search_fields=name%2Cdescription&order_by=name+A&return_count=True&skip=0&limit=10&user_id=1&shared=False")


@patch('geobox.model3d.urljoin')
@patch('geobox.model3d.urlencode')
def test_get_model(mock_urlencode, mock_urljoin, api, mock_model_data):
    api.get.return_value = mock_model_data
    mock_urlencode.return_value = 'f=json'

    mock_urljoin.return_value = f"3dmodels/{mock_model_data['uuid']}/?f=json"
    model = Model.get_model(api, mock_model_data['uuid'])
    assert model.uuid == mock_model_data['uuid']
    assert model.data == mock_model_data
    api.get.assert_called_once_with(f"3dmodels/{mock_model_data['uuid']}/?f=json")


@patch('geobox.model3d.urljoin')
@patch('geobox.model3d.urlencode')
def test_get_model_by_name(mock_urlencode, mock_urljoin, api, mock_model_data):
    api.get.return_value = [mock_model_data]
    mock_urlencode.return_value = 'f=json'

    mock_urljoin.return_value = f"{Model.BASE_ENDPOINT}?f=json"
    model = Model.get_model_by_name(api, name=mock_model_data['name'])
    assert model.uuid == mock_model_data['uuid']
    assert model.data == mock_model_data
    api.get.assert_called_once_with(f"{Model.BASE_ENDPOINT}?f=json&q=name+%3D+%27Treemodel%27&return_count=False&skip=0&limit=10&shared=False")
    # not found
    model = Model.get_model_by_name(api, name='not_found')
    assert model == None


def test_update(api, mock_model_data):
    model = Model(api, uuid=mock_model_data['uuid'], data=mock_model_data)
    updated_data = {
        'name': 'Updated Name',
        'description': 'Updated Description',
        'settings': {
            'model_settings': {
                'scale': 2.0,
                'rotation': [45, 0, 0],
                'location': [1, 1, 1]
            }
        },
        'thumbnail': 'https://example.com/new-thumbnail.jpg'
    }
    api.put.return_value = updated_data

    response = model.update(
        name='Updated Name',
        description='Updated Description',
        settings=updated_data['settings'],
        thumbnail='https://example.com/new-thumbnail.jpg'
    )
    assert response == updated_data
    assert model.name == 'Updated Name'
    assert model.description == 'Updated Description'
    api.put.assert_called_once_with(model.endpoint, updated_data)


def test_delete(api, mock_model_data):
    model = Model(api, uuid=mock_model_data['uuid'], data=mock_model_data)
    endpoint = model.endpoint
    model.delete()
    assert model.uuid is None
    assert model.endpoint is None
    api.delete.assert_called_once_with(endpoint)


def test_get_save_path_variants(api, mock_model_data, tmp_path):
    """Test _get_save_path edge cases."""
    model = Model(api, uuid=mock_model_data['uuid'], data=mock_model_data)

    # Case: no save_path -> returns cwd
    result = get_save_path()
    assert result == os.getcwd() + '/'

    # Case: save_path ends with '/' -> returns unchanged
    path = str(tmp_path) + "/"
    assert get_save_path(save_path=path) == path

    # Case: save_path does NOT end with '/' -> raises ValueError
    path_invalid = str(tmp_path)
    with pytest.raises(ValueError, match="save_path must end with a '/'"):
        get_save_path(save_path=path_invalid)


def test_create_progress_bar_import_error(api, mock_model_data):
    """Test _create_progress_bar when tqdm is missing."""
    model = Model(api, uuid=mock_model_data['uuid'], data=mock_model_data)
    model.size = 1234

    import builtins
    original_import = builtins.__import__

    def mocked_import(name, *args, **kwargs):
        if name == "tqdm.auto":
            raise ImportError("No module named 'tqdm'")
        return original_import(name, *args, **kwargs)

    with patch("builtins.__import__", side_effect=mocked_import):
        with patch("geobox.api.logger") as mock_logger:
            pbar = model._create_progress_bar()
            assert pbar is None
            mock_logger.warning.assert_called_once()


def test_download_with_obj(api, mock_model_data, tmp_path):
    """Test download() when self.data['obj'] exists."""
    # Make mock data have 'obj'
    mock_model_data['obj'] = "object-id"
    model = Model(api, uuid=mock_model_data['uuid'], data=mock_model_data)
    model.name = "test_model"
    model.endpoint = "models"

    # Mock api.get_model() to return a model with a normal download
    downloaded_model = MagicMock()
    downloaded_model.api = api
    downloaded_model.name = model.name
    downloaded_model.endpoint = model.endpoint
    downloaded_model.data = {}
    downloaded_model._get_save_path = lambda sp: str(tmp_path) + "/"
    downloaded_model._create_progress_bar = lambda: MagicMock()

    api.get_model.return_value = downloaded_model

    # Mock API response
    mock_response = MagicMock()
    mock_response.__enter__.return_value.iter_content.return_value = [b"chunk"]
    mock_response.__enter__.return_value.status_code = 200
    mock_response.__exit__.return_value = None

    with patch.object(api.session, "get", return_value=mock_response), \
         patch("builtins.open", MagicMock()), \
         patch("os.makedirs", return_value=None), \
         patch("os.remove", return_value=None), \
         patch("zipfile.ZipFile", MagicMock()):
        zip_path = model.download()
        assert zip_path.endswith(".zip")


def test_download_non_200(api, mock_model_data, tmp_path):
    """Test download() raises ApiRequestError on bad status code."""
    model = Model(api, uuid=mock_model_data['uuid'], data=mock_model_data)
    model.name = "test_model"
    model.endpoint = "models"

    mock_response = MagicMock()
    mock_response.__enter__.return_value.status_code = 404
    mock_response.__enter__.return_value.iter_content.return_value = []
    mock_response.__exit__.return_value = None

    with patch.object(api.session, "get", return_value=mock_response), \
         patch("os.makedirs", return_value=None):
        from geobox.api import ApiRequestError
        with pytest.raises(ApiRequestError, match="Failed to get model content: 404"):
            model.download(save_path=str(tmp_path)+"/")


def test_download_missing_uuid(api, mock_model_data):
    """Raise ValueError if UUID is missing."""
    model = Model(api, uuid=None, data=mock_model_data)  # uuid is None
    model.name = "test_model"
    model.endpoint = "models"

    from geobox.api import ApiRequestError

    with pytest.raises(ValueError, match="Model UUID is required to download content"):
        model.download()


def test_create_progress_bar_normal(api, mock_model_data):
    """Test _create_progress_bar returns tqdm instance when tqdm is available."""
    from tqdm.auto import tqdm

    model = Model(api, uuid=mock_model_data['uuid'], data=mock_model_data)
    model.size = 1024  # required for total=int(self.size)

    pbar = model._create_progress_bar()
    assert isinstance(pbar, tqdm)
    assert pbar.total == 1024
        

def test_thumbnail(api, mock_model_data):
    model = Model(api, uuid=mock_model_data['uuid'], data=mock_model_data)
    assert model.thumbnail == f"{api.base_url}{model.endpoint}thumbnail.png"


def test_share(api, mock_model_data, mock_user_data):
    """Test file sharing."""
    with patch.object(api, 'post'):
        model = Model(api, uuid=mock_model_data['uuid'], data=mock_model_data)
        users = [
            User(api, user_id=mock_user_data['1']['id'], data=mock_user_data['1']),
            User(api, user_id=mock_user_data['2']['id'], data=mock_user_data['2'])
        ]
        model.share(users=users)
        api.post.assert_called_once_with(
            f'{model.endpoint}share/',
            {'user_ids': [1, 2]}, is_json=False
        )

def test_unshare(api, mock_model_data, mock_user_data):
    """Test file unsharing."""
    with patch.object(api, 'post'):
        model = Model(api, uuid=mock_model_data['uuid'], data=mock_model_data)
        users = [
            User(api, user_id=mock_user_data['1']['id'], data=mock_user_data['1']),
            User(api, user_id=mock_user_data['2']['id'], data=mock_user_data['2'])
        ]
        model.unshare(users=users)
        api.post.assert_called_once_with(
            f'{model.endpoint}unshare/',
            {'user_ids': [1, 2]}, is_json=False
        )

def test_get_shared_users(api, mock_model_data, mock_user_data):
    """Test getting shared users."""
    mock_response = [
        mock_user_data['1'],
        mock_user_data['2']
    ]
    
    with patch.object(api, 'get', return_value=mock_response):
        model = Model(api, uuid=mock_model_data['uuid'], data=mock_model_data)
        result = model.get_shared_users(search='user', limit=2)

        api.get.assert_called_once_with(
            f'{model.endpoint}shared-with-users/?search=user&skip=0&limit=2'
        )

        assert len(result) == 2
        assert result[0].first_name == 'test 1'
        assert result[0].last_name == 'test 1'
        assert result[1].first_name == 'test 2'
        assert result[1].last_name == 'test 2'


def test_to_async(api, async_api, mock_model_data):
    model = Model(api, uuid=mock_model_data['uuid'], data=mock_model_data)
    async_instance = model.to_async(async_api)
    assert async_instance.api == async_api 


import pytest
from unittest.mock import patch, ANY
from urllib.parse import urljoin, urlencode

from geobox.user import User
from geobox.enums import UserRole, UserStatus
from geobox.plan import Plan
from geobox.utils import xor_encode


def test_init(api, mock_user_data):
    user = User(api, user_id=mock_user_data['1']['id'], data=mock_user_data['1'])
    assert user.data == mock_user_data['1']
    assert user.endpoint == f"users/{mock_user_data['1']['id']}/"
    assert user.api == api


def test_repr(api, mock_user_data):
    user = User(api, user_id=mock_user_data['1']['id'], data=mock_user_data['1'])
    assert repr(user) == f"User(id={user.id}, first_name=test 1, last_name=test 1)"


def test_role(api, mock_admin_user_data):
    user = User(api, user_id=mock_admin_user_data['id'], data=mock_admin_user_data)
    assert user.role == UserRole.ACCOUNT_ADMIN


def test_status(api, mock_admin_user_data):
    user = User(api, user_id=mock_admin_user_data['id'], data=mock_admin_user_data)
    assert user.status == UserStatus.DISABLED


def test_plan(api, mock_admin_user_data):
    user = User(api, user_id=mock_admin_user_data['id'], data=mock_admin_user_data)
    assert type(user.plan) == Plan

def test_get_users(api, mock_admin_user_data):
    api.get.return_value = [mock_admin_user_data, mock_admin_user_data]
    result = User.get_users(api, search='test')
    assert isinstance(result, list)
    assert isinstance(result[0], User)
    api.get.assert_called_once_with('users/?f=json&search=test&return_count=False&skip=0&limit=10')

def test_create_user(api, mock_admin_user_data):
    api.post.return_value = mock_admin_user_data
    user = User.create_user(
        api,
        username=mock_admin_user_data['username'],
        email=mock_admin_user_data['email'],
        password='password',
        role=UserRole.ACCOUNT_ADMIN,
        first_name=mock_admin_user_data['first_name'],
        last_name=mock_admin_user_data['last_name'],
        mobile=mock_admin_user_data['mobile'],
        status=UserStatus.ACTIVE
    )
    assert isinstance(user, User)
    expected_data = {
        "username": mock_admin_user_data['username'],
        "email": mock_admin_user_data['email'],
        "password": xor_encode('password'),
        "role": UserRole.ACCOUNT_ADMIN.value,
        "first_name": mock_admin_user_data['first_name'],
        "last_name": mock_admin_user_data['last_name'],
        "status": UserStatus.ACTIVE.value
    }
    api.post.assert_called_once_with(User.BASE_ENDPOINT, expected_data)


def test_search_users(api, mock_user_data):
    api.get.return_value = [mock_user_data['1'], mock_user_data['1']]
    result = User.search_users(api, search='Test', skip=1, limit=5)
    assert isinstance(result, list)
    assert isinstance(result[0], User)
    api.get.assert_called_once_with(f"{User.BASE_ENDPOINT}search/?search=Test&skip=1&limit=5")


def test_get_user(api, mock_admin_user_data):
    api.get.return_value = mock_admin_user_data
    user = User.get_user(api, user_id=mock_admin_user_data['id'])
    assert isinstance(user, User)
    expected_params = {'f': 'json'}
    api.get.assert_called_once_with(f"{User.BASE_ENDPOINT}{mock_admin_user_data['id']}/?f=json")


def test_update(api, mock_admin_user_data):
    user = User(api, user_id=mock_admin_user_data['id'], data=mock_admin_user_data)
    updated = mock_admin_user_data.copy()
    updated['status'] = UserStatus.PENDING.value
    api.put.return_value = updated
    result = user.update(status=UserStatus.PENDING)
    api.put.assert_called_once_with(f"{user.endpoint}", {'status': 'Pending'})
    assert result['status'] == UserStatus.PENDING.value


def test_delete(api, mock_admin_user_data):
    user = User(api, user_id=mock_admin_user_data['id'], data=mock_admin_user_data)
    endpoint = user.endpoint
    user.delete()
    api.delete.assert_called_once_with(endpoint)
    assert user.uuid is None
    assert user.endpoint is None


def test_get_sessions(api, mock_admin_user_data):
    user = User(api, user_id=mock_admin_user_data['id'], data=mock_admin_user_data)
    with patch.object(api, 'get', return_value=[{'uuid': 'sess-1', 'id': 1}, {'uuid': 'sess-2', 'id': 2}]):
        sessions = user.get_sessions()
        assert len(sessions) == 2
        assert sessions[0].uuid == 'sess-1'
        assert sessions[1].uuid == 'sess-2'
        expected_endpoint = urljoin(user.endpoint, 'sessions/?f=json')
        api.get.assert_called_once_with(expected_endpoint) 

    with patch.object(api, 'get', return_value=[{'uuid': 'sess-1', 'id': 1}, {'uuid': 'sess-2', 'id': 2}]) as mock_api,\
         patch.object(User, 'get_user', return_value=User(api, mock_admin_user_data['id'], mock_admin_user_data)) as mock_get_user:
        sessions = user.get_sessions(user_id=1)
        assert len(sessions) == 2
        assert sessions[0].uuid == 'sess-1'
        assert sessions[1].uuid == 'sess-2'
        expected_endpoint = f"{user.BASE_ENDPOINT}1/sessions/?f=json"
        api.get.assert_called_once_with(expected_endpoint) 


def test_change_password(api, mock_admin_user_data):
    user = User(api, user_id=mock_admin_user_data['id'], data=mock_admin_user_data)
    with patch.object(api, 'post', return_value={"message": "success"}):
        user.change_password('new_password')
        expected_endpoint = urljoin(user.endpoint, 'change-password')
        api.post.assert_called_once_with(expected_endpoint, {'new_password': xor_encode('new_password')}, is_json=False)


def test_renew_plan(api, mock_admin_user_data):
    user = User(api, user_id=mock_admin_user_data['id'], data=mock_admin_user_data)
    with patch.object(api, 'post'):
        user.renew_plan()
        expected_endpoint = urljoin(user.endpoint, 'renewPlan')
        api.post.assert_called_once_with(expected_endpoint)


def test_user_to_async(api, async_api, mock_admin_user_data):
    user = User(api, user_id=mock_admin_user_data['id'], data=mock_admin_user_data)
    async_instance = user.to_async(async_api)
    assert async_instance.api == async_api


def test_session_repr(api, mock_admin_user_data, mock_session_data):
    user = User(api, user_id=mock_admin_user_data['id'], data=mock_admin_user_data)
    api.get.return_value = [mock_session_data, mock_session_data]
    sessions = user.get_sessions()
    assert repr(sessions[0]) == f"Session(user={user}, agent='{sessions[0].agent}')"


def test_close_seesion(api, mock_admin_user_data, mock_session_data):
    user = User(api, user_id=mock_admin_user_data['id'], data=mock_admin_user_data)
    api.get.return_value = [mock_session_data, mock_session_data]
    sessions = user.get_sessions()
    api.reset_mock()
    sessions[0].close()
    api.post.assert_called_once_with('users/165/sessions/3857eb38-69d8-4b3c-8f60-29bac352a8d9', {'user_id': 165, 'session_uuid': '3857eb38-69d8-4b3c-8f60-29bac352a8d9'})


def test_session_to_async(api, async_api, mock_admin_user_data, mock_session_data):
    user = User(api, user_id=mock_admin_user_data['id'], data=mock_admin_user_data)
    api.get.return_value = [mock_session_data, mock_session_data]
    sessions = user.get_sessions()
    async_instance = sessions[0].to_async(async_api)
    assert async_instance.user.api == async_api


from urllib.parse import urljoin
from typing import Optional, List, Dict, Any, TYPE_CHECKING

from .base import AsyncBase
from ..enums import FeatureType

if TYPE_CHECKING:
    from .vectorlayer import VectorLayer
    from ..api import GeoboxClient
    from ..feature import Feature


class AsyncFeature(AsyncBase):

    BASE_SRID = 3857

    def __init__(self, 
        layer: 'VectorLayer', 
        srid: Optional[int] = 3857,
        data: Optional[Dict] = {}):
        """
        Constructs all the necessary attributes for the Feature object.

        Args:
            layer (VectorLayer): The vector layer this feature belongs to
            srid (int, optional): The Spatial Reference System Identifier (default is 3857)
            data (Dict, optional): The feature data contains the feature geometry and properties
        """
        super().__init__(api=layer.api)
        self.layer = layer
        self._srid = srid
        self.data = data or {}
        self.original_geometry = self.data.get('geometry')
        self.endpoint = urljoin(layer.endpoint, f'features/{self.data.get("id")}/') if self.data.get('id') else None


    def __dir__(self) -> List[str]:
        """
        Return a list of available attributes for the Feature object.
        
        This method extends the default dir() behavior to include:
        - All keys from the feature data dictionary
        - All keys from the geometry dictionary
        - All keys from the properties dictionary
        
        This allows for better IDE autocompletion and introspection of feature attributes.
        
        Returns:
            list: A list of attribute names available on this Feature object.
        """
        return super().__dir__() + list(self.data.keys()) + list(self.data.get('geometry').keys()) + list(self.data.get('properties').keys())


    def __repr__(self) -> str:
        """
        Return a string representation of the Feature object.

        Returns:
            str: A string representation of the Feature object.
        """
        return f"AsyncFeature(id={self.id}, type={self.feature_type})"
    

    def __getattr__(self, name: str) -> Any:
        """
        Get an attribute from the resource.

        Args:
            name (str): The name of the attribute
        """
        if name in self.data:
            return self.data.get(name)
        # elif name in self.data['geometry']:
        #     return self.data['geometry'].get(name)
        elif name in self.data['properties']:
            return self.data['properties'].get(name)
        
        raise AttributeError(f"Feature has no attribute {name}")


    @property
    def srid(self) -> int:
        """
        Get the Spatial Reference System Identifier (SRID) of the feature.

        Returns:
            int: The SRID of the feature.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>>     feature = await layer.get_feature(id=1)
            >>>     feature.srid  # 3857
        """
        return self._srid
    

    @property
    def feature_type(self) -> 'FeatureType':
        """
        Get the type of the feature.

        Returns:
            FeatureType: The type of the feature.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>>     feature = await layer.get_feature(id=1)
            >>>     feature.feature_type
        """
        return FeatureType(self.data.get('geometry').get('type')) if self.data.get('geometry') else None
    

    @property
    def coordinates(self) -> List[float]:
        """
        Get the coordinates of the ferepoature.

        Returns:
            list: The coordinates of the feature.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>>     feature = await layer.get_feature(id=1)
            >>>     feature.coordinates
        """
        return self.data.get('geometry').get('coordinates') if self.data.get('geometry') else None
    

    @coordinates.setter
    def coordinates(self, value: List[float]) -> None:
        """
        Set the coordinates of the feature.

        Args:
            value (list): The coordinates to set.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>>     feature = await layer.get_feature(id=1)
            >>>     feature.coordinates = [10, 20]
        """
        self.data['geometry']['coordinates'] = value
    

    @property
    async def length(self) -> float:
        """
        [async] Returns the length of thefeature geometry (geometry package extra is required!)

        Returns:
            float: the length of thefeature geometry

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>>     feature = await layer.get_feature(id=1)
            >>>     await feature.length
        """
        try:
            return self.geom_length
        except AttributeError:
            endpoint = f'{self.endpoint}length'
            return await self.api.get(endpoint)


    @property
    async def area(self) -> float:
        """
        [async] Returns the area of thefeature geometry (geometry package extra is required!)

        Returns:
            float: the area of thefeature geometry

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>>     feature = await layer.get_feature(id=1)
            >>>     await feature.area
        """
        try:
            return self.geom_area
        except AttributeError:
            endpoint = f'{self.endpoint}area'
            return await self.api.get(endpoint)
    
    
    async def save(self) -> None:
        """
        [async] Save the feature. Creates a new feature if feature_id is None, updates existing feature otherwise.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>>     feature = await layer.get_feature(id=1)
            >>>     feature.properties['name'] = 'New Name'
            >>>     await feature.save()
        """
        data = self.data.copy()
        try:
            if self.id:
                if self.srid != self.BASE_SRID:
                    self.data['geometry'] = self.original_geometry
                await self.update(self.data)
                self.coordinates = data['geometry']['coordinates']
        except AttributeError:
            endpoint = urljoin(self.layer.endpoint, 'features/')
            response = await self.layer.api.post(endpoint, data)
            self.endpoint = urljoin(self.layer.endpoint, f'features/{response["id"]}/')        
            self.data.update(response)


    async def delete(self) -> None:
        """
        [async] Delete the feature.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>>     feature = await layer.get_feature(id=1)
            >>>     await feature.delete()
        """
        await super()._delete(self.endpoint)


    async def update(self, geojson: Dict) -> Dict:
        """
        [async] Update the feature data property.

        Args:
            geojson (Dict): The GeoJSON data for the feature

        Returns:
            Dict: The response from the API.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>>     feature = await layer.get_feature(id=1)
            >>>     geojson = {
            ...         "geometry": {
            ...             "type": "Point",
            ...             "coordinates": [10, 20]
            ...         }
            ...     }   
            >>>     await feature.update(geojson)
        """
        return await super()._update(self.endpoint, geojson, clean=False)
    

    @classmethod
    async def create_feature(cls, layer: 'VectorLayer', geojson: Dict) -> 'AsyncFeature':
        """
        [async] Create a new feature in the vector layer.

        Args:
            layer (VectorLayer): The vector layer to create the feature in
            geojson (Dict): The GeoJSON data for the feature

        Returns:
            AsyncFeature: The created feature instance

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>>     geojson = {
            ...        "type": "Feature",
            ...        "geometry": {"type": "Point", "coordinates": [10, 20]},
            ...        "properties": {"name": "My Point"}
            ...     }
            >>>     feature = await Feature.create_feature(layer, geojson)
        """
        endpoint = urljoin(layer.endpoint, 'features/')
        return await cls._create(layer.api, endpoint, geojson, factory_func=lambda api, item: AsyncFeature(layer, data=item))


    @classmethod
    async def get_feature(cls, layer: 'VectorLayer', feature_id: int, user_id: int = None) -> 'AsyncFeature':
        """
        [async] Get a feature by its ID.

        Args:
            layer (VectorLayer): The vector layer the feature belongs to
            feature_id (int): The ID of the feature
            user_id (int): specific user. privileges required.

        Returns:
            AsyncFeature: The retrieved feature instance

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.feature import AsyncFeature
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>>     feature = await AsyncFeature.get_feature(layer, feature_id=1)
        """
        param = {
            'f': 'json',
            'user_id': user_id
        }
        endpoint = urljoin(layer.endpoint, f'features/')
        return await cls._get_detail(layer.api, endpoint, uuid=feature_id, params=param, factory_func=lambda api, item: AsyncFeature(layer, data=item))


    @property
    def geometry(self) -> 'BaseGeometry':
        """
        Get the feature geometry as a Shapely geometry object.

        Returns:
            shapely.geometry.BaseGeometry: The Shapely geometry object representing the feature's geometry

        Raises:
            ValueError: If the geometry is not a dictionary
            ValueError: If the geometry type is not present in the feature data
            ValueError: If the geometry coordinates are not present in the feature data
            ImportError: If shapely is not installed

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>>     feature = await layer.get_feature(id=1)
            >>>     feature.geometry
        """
        try:
            from shapely.geometry import shape
        except ImportError:
            raise ImportError(
                "The 'geometry' extra is required for this function. "
                "Install it with: pip install geobox[geometry]"
            )
        
        if not self.data:
            return None
        
        elif not self.data.get('geometry'):
            raise ValueError("Geometry is not present in the feature data")
        
        elif not isinstance(self.data['geometry'], dict):
            raise ValueError("Geometry is not a dictionary")
        
        elif not self.data['geometry'].get('type'):
            raise ValueError("Geometry type is not present in the feature data")
        
        elif not self.data['geometry'].get('coordinates'):
            raise ValueError("Geometry coordinates are not present in the feature data")
        
        else:
            return shape(self.data['geometry'])
    

    @geometry.setter
    def geometry(self, value: object) -> None:
        """
        Set the feature geometry.

        Args:
            value (object): The geometry to set.

        Raises:
            ValueError: If geometry type is not supported
            ValueError: If the geometry has a different type than the layer type
            ImportError: If shapely is not installed

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from shapely.affinity import translate
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>>     feature = await layer.get_feature(id=1)
            >>>     geom = feature.geometry
            >>>     geom = translate(geom, 3.0, 0.5) # example change applied to the feature's geometry
            >>>     feature.geometry = geom
            >>>     await feature.save()
        """
        try:
            from shapely.geometry import mapping, Point, MultiPoint, LineString, MultiLineString, Polygon, MultiPolygon
        except ImportError:
            raise ImportError(
                "The 'geometry' extra is required for this function. "
                "Install it with: pip install geobox[geometry]"
            )
        
        if not isinstance(value, (Point, MultiPoint, LineString, MultiLineString, Polygon, MultiPolygon)):
            raise ValueError("Geometry must be a Shapely geometry object")
        
        elif self.feature_type and value.geom_type != self.feature_type.value:
            raise ValueError("Geometry must have the same type as the layer type")
        
        else:
            self.data['geometry'] = mapping(value)

        
    def transform(self, out_srid: int) -> 'AsyncFeature':
        """
        Transform the feature geometry to a new SRID.

        Args:
            out_srid (int): The target SRID to transform the geometry to (e.g., 4326 for WGS84, 3857 for Web Mercator)

        Returns:
            AsyncFeature: A new Feature instance with transformed geometry.

        Raises:
            ValueError: If the feature has no geometry or if the transformation fails.
            ImportError: If pyproj is not installed.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>>     feature = await layer.get_feature(id=1, srid=3857)
            >>>     # Transform from Web Mercator (3857) to WGS84 (4326)
            >>>     transformed = feature.transform(out_srid=4326)
            >>>     transformed.srid  # 4326
        """
        try:
            from pyproj import Transformer
            from shapely.geometry import Point, LineString, Polygon, MultiPoint, MultiLineString, MultiPolygon, mapping
        except ImportError:
            raise ImportError(
                "The 'geometry' extra is required for this function. "
                "Install it with: pip install geobox[geometry]"
            )

        if not self.data or not self.data.get('geometry'):
            raise ValueError("Feature geometry is required for transformation")

        # Get the current SRID from the feature or default to 3857 (Web Mercator)
        current_srid = self.srid or 3857

        # Create transformer
        transformer = Transformer.from_crs(current_srid, out_srid, always_xy=True)

        # Get the geometry
        geom = self.geometry

        # Transform coordinates based on geometry type
        if geom.geom_type == 'Point':
            x, y = geom.x, geom.y
            new_x, new_y = transformer.transform(x, y)
            new_geom = Point(new_x, new_y)
        elif geom.geom_type == 'LineString':
            coords = list(geom.coords)
            new_coords = [transformer.transform(x, y) for x, y in coords]
            new_geom = LineString(new_coords)
        elif geom.geom_type == 'Polygon':
            exterior = [transformer.transform(x, y) for x, y in geom.exterior.coords]
            interiors = [[transformer.transform(x, y) for x, y in interior.coords] for interior in geom.interiors]
            new_geom = Polygon(exterior, holes=interiors)
        elif geom.geom_type == 'MultiPoint':
            new_geoms = [Point(transformer.transform(point.x, point.y)) for point in geom.geoms]
            new_geom = MultiPoint(new_geoms)
        elif geom.geom_type == 'MultiLineString':
            new_geoms = [LineString([transformer.transform(x, y) for x, y in line.coords]) for line in geom.geoms]
            new_geom = MultiLineString(new_geoms)
        elif geom.geom_type == 'MultiPolygon':
            new_geoms = []
            for poly in geom.geoms:
                exterior = [transformer.transform(x, y) for x, y in poly.exterior.coords]
                interiors = [[transformer.transform(x, y) for x, y in interior.coords] for interior in poly.interiors]
                new_geoms.append(Polygon(exterior, holes=interiors))
            new_geom = MultiPolygon(new_geoms)

        # update the feature data
        self.data['geometry'] = mapping(new_geom)
        self._srid = out_srid

        return self


    def to_sync(self, sync_client: 'GeoboxClient') -> 'Feature':
        """
        [async] Switch to sync version of the feature instance to have access to the sync methods

        Args:
            sync_client (GeoboxClient): The sync version of the GeoboxClient instance for making requests.

        Returns:
            Feature: the sync instance of the feature.

        Example:
            >>> from geobox import Geoboxclient
            >>> from geobox.aio import AsyncGeoboxClient
            >>> client = GeoboxClient()
            >>> async with AsyncGeoboxClient() as async_client:
            >>>     layer = await async_client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>>     feature = await layer.get_feature(id=1, srid=3857)
            >>>     sync_feature = feature.to_sync(client)
        """
        from ..feature import Feature

        sync_layer = self.layer.to_sync(sync_client=sync_client)
        return Feature(layer=sync_layer, srid=self.srid, data=self.data)


from urllib.parse import urljoin, urlencode
from typing import Dict, List, Optional, Union, TYPE_CHECKING

from .utils import clean_data
from .raster import Raster

if TYPE_CHECKING:
    from . import GeoboxClient
    from .user import User
    from .task import Task
    from .aio import AsyncGeoboxClient
    from .aio.mosaic import AsyncMosaic


class Mosaic(Raster):

    BASE_ENDPOINT: str = 'mosaics/'

    def __init__(self, 
                 api: 'GeoboxClient', 
                 uuid: str, 
                 data: Optional[Dict] = {}):
        """
        Initialize a Mosaic instance.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            uuid (str): The unique identifier for the mosaic.
            data (Dict, optional): The data of the mosaic.
        """
        super().__init__(api, uuid, data)
    

    @classmethod
    def get_mosaics(cls, api: 'GeoboxClient', **kwargs) -> Union[List['Mosaic'], int]:
        """
        Get a list of mosaics.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.

        Keyword Args:
            q (str): query filter based on OGC CQL standard. e.g. "field1 LIKE '%GIS%' AND created_at > '2021-01-01'".
            seacrh (str): search term for keyword-based searching among search_fields or all textual fields if search_fields does not have value. NOTE: if q param is defined this param will be ignored.
            search_fields (str): comma separated list of fields for searching.
            order_by (str): comma separated list of fields for sorting results [field1 A|D, field2 A|D, …]. e.g. name A, type D. NOTE: "A" denotes ascending order and "D" denotes descending order.
            return_count (bool): if true, the number of mosaics will be returned.
            skip (int): number of mosaics to skip. minimum value is 0.
            limit (int): maximum number of mosaics to return. minimum value is 1.
            user_id (int): specific user. privileges required.
            shared (bool): Whether to return shared mosaics. default is False.

        Returns:
            List['Mosaic'] | int: A list of Mosaic instances or the number of mosaics.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.mosaic import Mosaic
            >>> client = GeoboxClient()
            >>> mosaics = Mosaic.get_mosaics(client, q="name LIKE '%GIS%'")
            or
            >>> mosaics = client.get_mosaics(q="name LIKE '%GIS%'")
        """
        params = {
            'terrain': kwargs.get('terrain', None),
            'f': 'json',
            'q': kwargs.get('q', None),
            'search': kwargs.get('search', None),
            'search_fields': kwargs.get('search_fields', None),
            'order_by': kwargs.get('order_by', None),
            'return_count': kwargs.get('return_count', False),
            'skip': kwargs.get('skip', 0),
            'limit': kwargs.get('limit', 100),
            'user_id': kwargs.get('user_id', None),
            'shared': kwargs.get('shared', False)
        }
        return super()._get_list(api, cls.BASE_ENDPOINT, params, factory_func=lambda api, item: Mosaic(api, item['uuid'], item))
    
    
    @classmethod
    def get_mosaics_by_ids(cls, api: 'GeoboxClient', ids: List[str], user_id: int = None) -> List['Mosaic']:
        """
        Get mosaics by their IDs.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            ids (List[str]): The IDs of the mosaics.
            user_id (int, optional): specific user. privileges required.

        Returns:
            List[Mosaic]: A list of Mosaic instances.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.mosaic import Mosaic
            >>> client = GeoboxClient()
            >>> mosaics = Mosaic.get_mosaics_by_ids(client, ids=['1, 2, 3'])
            or
            >>> mosaics = client.get_mosaics_by_ids(ids=['1, 2, 3'])
        """
        params = {
            'ids': ids,
            'user_id': user_id
        }
        endpoint = urljoin(cls.BASE_ENDPOINT, 'get-mosaics/')
        return super()._get_list_by_ids(api, endpoint, params, factory_func=lambda api, item: Mosaic(api, item['uuid'], item))

    
    @classmethod
    def create_mosaic(cls,
                      api: 'GeoboxClient',
                      name:str,
                      display_name: str = None,
                      description: str = None,
                      pixel_selection: str = None,
                      min_zoom: int = None,
                      user_id: int = None) -> 'Mosaic':
        """
        Create New Raster Mosaic

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            name (str): The name of the mosaic.
            display_name (str, optional): The display name of the mosaic.
            description (str, optional): The description of the mosaic.
            pixel_selection (str, optional): The pixel selection of the mosaic.
            min_zoom (int, optional): The minimum zoom of the mosaic.
            user_id (int, optional): specific user. privileges required.

        Returns:
            Mosaic: The created mosaic.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.mosaic import Mosaic
            >>> client = GeoboxClient()
            >>> mosaic = Mosaic.create_mosaic(client, name='mosaic_name')
            or
            >>> mosaic = client.create_mosaic(name='mosaic_name')
        """
        data = {
            'name': name,
            'display_name': display_name,
            'description': description,
            'pixel_selection': pixel_selection,
            'min_zoom': min_zoom,
            'user_id': user_id
        }
        return super()._create(api, cls.BASE_ENDPOINT, data, factory_func=lambda api, item: Mosaic(api, item['uuid'], item))
    

    @classmethod
    def get_mosaic(cls, api: 'GeoboxClient', uuid: str, user_id: int = None) -> 'Mosaic':
        """
        Get a mosaic by uuid.

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            uuid (str): The UUID of the mosaic.
            user_id (int, optional): specific user. privileges required.

        Returns:
            Mosaic: The mosaic object.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.mosaic import Mosaic
            >>> client = GeoboxClient()
            >>> mosaic = Mosaic.get_mosaic(client, uuid="12345678-1234-5678-1234-567812345678")
            or
            >>> mosaic = client.get_mosaic(uuid="12345678-1234-5678-1234-567812345678")
        """      
        params = {
            'f': 'json',
            'user_id': user_id
        }
        return super()._get_detail(api, cls.BASE_ENDPOINT, uuid, params, factory_func=lambda api, item: Mosaic(api, item['uuid'], item))


    @classmethod
    def get_mosaic_by_name(cls, api: 'GeoboxClient', name: str, user_id: int = None) -> Union['Mosaic', None]:
        """
        Get a mosaic by name

        Args:
            api (GeoboxClient): The GeoboxClient instance for making requests.
            name (str): the name of the mosaic to get
            user_id (int, optional): specific user. privileges required.

        Returns:
            Mosaic | None: returns the mosaic if a mosaic matches the given name, else None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.mosaic import Mosaic
            >>> client = GeoboxClient()
            >>> mosaic = Mosaic.get_mosaic_by_name(client, name='test')
            or
            >>> mosaic = client.get_mosaic_by_name(name='test')
        """
        mosaics = cls.get_mosaics(api, q=f"name = '{name}'", user_id=user_id)
        if mosaics and mosaics[0].name == name:
            return mosaics[0]
        else:
            return None


    def update(self, **kwargs) -> Dict:
        """
        Update a mosaic.

        Keyword Args:
            name (str): The name of the mosaic.
            display_name (str): The display name of the mosaic.
            description (str): The description of the mosaic.
            pixel_selection (str): The pixel selection of the mosaic.
            min_zoom (int): The minimum zoom of the mosaic.

        Returns:
            Dict: the updated data

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.mosaic import Mosaic
            >>> client = GeoboxClient()
            >>> mosaic = Mosaic.get_mosaic(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> mosaic.update(name='new_name', display_name='new_display_name', description='new_description', pixel_selection='new_pixel_selection', min_zoom=10)
        """
        data = {
            'name': kwargs.get('name'),
            'display_name': kwargs.get('display_name'),
            'description': kwargs.get('description'),
            'pixel_selection': kwargs.get('pixel_selection'),
            'min_zoom': kwargs.get('min_zoom')
        }
        return super()._update(self.endpoint, data)


    def delete(self) -> None:
        """
        Delete the mosaic.

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.mosaic import Mosaic
            >>> client = GeoboxClient()
            >>> mosaic = Mosaic.get_mosaic(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> mosaic.delete()
        """
        super()._delete(self.endpoint)
    

    @property
    def thumbnail(self) -> str:
        """
        Get the thumbnail of the mosaic.

        Returns:
            str: The thumbnail url of the mosaic.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.mosaic import Mosaic
            >>> client = GeoboxClient()
            >>> mosaic = Mosaic.get_mosaic(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> mosaic.thumbnail
        """
        return super()._thumbnail(format="")


    def get_point(self, lat: float, lng: float) -> List[float]:
        """
        Get the points of the mosaic.

        Args:
            lat (float): The latitude of the point.
            lng (float): The longitude of the point.

        Returns:
            List[float]: The points of the mosaic.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.mosaic import Mosaic
            >>> client = GeoboxClient()
            >>> mosaic = Mosaic.get_mosaic(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> mosaic.get_point(lat=60, lng=50)
        """
        return super().get_point(lat, lng)


    def get_render_png_url(self, x: int = '{x}', y: int = '{y}', z: int = '{z}', **kwargs) -> str:
        """
        Get the tile render URL of the mosaic.

        Args:
            x (int, optional): The x coordinate of the tile.
            y (int, optional): The y coordinate of the tile.
            z (int, optional): The zoom level of the tile.

        Keyword Args:
            indexes (str, optional): list of comma separated band indexes to be rendered. e.g. 1, 2, 3
            nodata (int, optional)
            expression (str, optional): band math expression. e.g. b1*b2+b3
            rescale (List, optional): comma (',') separated Min,Max range. Can set multiple time for multiple bands.
            color_formula (str, optional): Color formula. e.g. gamma R 0.5
            colormap_name (str, optional)
            colormap (str, optional): JSON encoded custom Colormap. e.g. {"0": "#ff0000", "1": "#00ff00"} or [[[0, 100], "#ff0000"], [[100, 200], "#00ff00"]]

        Returns:
            str: The tile render URL of the mosaic.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.mosaic import Mosaic
            >>> client = GeoboxClient()
            >>> mosaic = Mosaic.get_mosaic(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> mosaic.get_tile_render_url(x=1, y=1, z=1)
        """
        return super().get_render_png_url(x, y, z, **kwargs)
    

    def get_tile_png_url(self, x: int = '{x}', y: int = '{y}', z: int = '{z}') -> str:
        """
        Get the tile PNG URL of the mosaic.

        Args:
            x (int, optional): The x coordinate of the tile.
            y (int, optional): The y coordinate of the tile.
            z (int, optional): The zoom level of the tile.

        Returns:
            str: The tile PNG URL of the mosaic.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.mosaic import Mosaic
            >>> client = GeoboxClient()
            >>> mosaic = Mosaic.get_mosaic(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> mosaic.get_tile_png_url(x=1, y=1, z=1)
        """
        return super().get_tile_png_url(x, y, z)
    

    def get_tile_json(self) -> str:
        """
        Get the tile JSON of the raster.

        Returns:
            str: The tile JSON of the raster.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.mosaic import Mosaic
            >>> client = GeoboxClient()
            >>> mosaic = Mosaic.get_mosaic(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> mosaic.get_tile_json()
        """
        return super().get_tile_json()
    

    def wmts(self, scale: int = None) -> str:
        """
        Get the WMTS URL

        Args:
            scale (int, optional): The scale of the raster. values are: 1, 2

        Returns:
            str: The WMTS URL of the mosaic.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.mosaic import Mosaic
            >>> client = GeoboxClient()
            >>> mosaic = Mosaic.get_mosaic(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> mosaic.wmts(scale=1)
        """
        return super().wmts(scale)

    @property
    def settings(self) -> Dict:
        """
        Get the settings of the mosaic.

        Returns:
            Dict: The settings of the mosaic.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.mosaic import Mosaic
            >>> client = GeoboxClient()
            >>> mosaic = Mosaic.get_mosaic(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> mosaic.settings
        """
        return super().settings
    

    def update_settings(self, settings: Dict) -> Dict:
        """
        Update the settings

        settings (Dict): settings dictionary

        Returns:
            Dict: updated settings

        Example:
            >>> from geobox import GeoboxClient
            >>> client = GeoboxClient()
            >>> mosaic1 = client.get_mosaic(uuid="12345678-1234-5678-1234-567812345678")
            >>> mosaic2 = client.get_mosaic(uuid="12345678-1234-5678-1234-567812345678")
            >>> mosaic1.update_settings(mosaic2.settings)
        """
        return super().update_settings(settings)  
    
    
    def set_settings(self, **kwargs) -> None:
        """
        Set the settings of the mosaic.

        Keyword Args:
            nodata (int): The nodata value of the raster.
            indexes (list[int]): The indexes of the raster.
            rescale (list[int]): The rescale of the raster.
            colormap_name (str): The colormap name of the raster.
            color_formula (str): The color formula of the raster.
            expression (str): The expression of the raster.
            exaggeraion (int): The exaggeraion of the raster.
            min_zoom (int): The min zoom of the raster.
            max_zoom (int): The max zoom of the raster.
            use_cache (bool): Whether to use cache of the raster.
            cache_until_zoom (int): The cache until zoom of the raster.

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.mosaic import Mosaic
            >>> client = GeoboxClient()
            >>> mosaic = Mosaic.get_mosaic(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> mosaic.set_settings(nodata=0, 
            ...                         indexes=[1], 
            ...                         rescale=[[0, 10000]], 
            ...                         colormap_name='gist_rainbow', 
            ...                         color_formula='Gamma R 0.5', 
            ...                         expression='b1 * 2', 
            ...                         exaggeraion=10, 
            ...                         min_zoom=0, 
            ...                         max_zoom=22, 
            ...                         use_cache=True, 
            ...                         cache_until_zoom=17)
        """
        return super().set_settings(**kwargs)


    def get_rasters(self, user_id: int = None) -> List[Raster]:
        """
        Get the rasters of the mosaic

        Args:
            user_id (int, optional): specific user. privileges required.

        Returns:
            List[Raster]: The rasters of the mosaic.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.mosaic import Mosaic
            >>> client = GeoboxClient()
            >>> mosaic = Mosaic.get_mosaic(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> rasters = mosaic.get_rasters()
        """
        params = clean_data({
            'user_id': user_id
        })
        query_string = urlencode(params)
        endpoint = urljoin(self.BASE_ENDPOINT, f'{self.uuid}/rasters?{query_string}')
        response = self.api.get(endpoint)
        return [Raster(self.api, raster_data['uuid'], raster_data) for raster_data in response]
    

    def add_rasters(self, rasters: List['Raster']) -> None:
        """
        Add a raster to the mosaic.

        Args:
            rasters (List[Raster]): list of raster objects to add

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.mosaic import Mosaic
            >>> client = GeoboxClient()
            >>> mosaic = Mosaic.get_mosaic(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> rasters = client.get_rasters()
            >>> mosaic.add_raster(rasters=rasters)
        """
        data = clean_data({
            'raster_ids': [raster.id for raster in rasters]
        })
        endpoint = urljoin(self.BASE_ENDPOINT, f'{self.uuid}/rasters/')
        self.api.post(endpoint, data, is_json=False)
    

    def remove_rasters(self, rasters: List['Raster']) -> None:
        """
        Remove a raster from the mosaic.

        Args:
            rasters (List[Raster]): list of raster objects to remove

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.mosaic import Mosaic
            >>> client = GeoboxClient()
            >>> mosaic = Mosaic.get_mosaic(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> rasters = client.get_raster()
            >>> mosaic.remove_rasters(rasters=rasters)
        """
        param = clean_data({
            'raster_ids': [raster.id for raster in rasters]
        })
        query_string = urlencode(param)
        endpoint = urljoin(self.BASE_ENDPOINT, f'{self.uuid}/rasters/?{query_string}')
        self.api.delete(endpoint, is_json=False)


    def share(self, users: List['User']) -> None:
        """
        Shares the mosaic with specified users.

        Args:
            users (List[User]): The list of user objects to share the mosaic with.

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.mosaic import Mosaic
            >>> client = GeoboxClient()
            >>> mosaic = Mosaic.get_mosaic(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> users = client.search_users(search='John')
            >>> mosaic.share(users=users)
        """
        super()._share(self.endpoint, users)
    

    def unshare(self, users: List['User']) -> None:
        """
        Unshares the mosaic with specified users.

        Args:
            users (List[User]): The list of user objects to unshare the mosaic with.

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.mosaic import Mosaic
            >>> client = GeoboxClient()
            >>> mosaic = Mosaic.get_mosaic(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> users = client.search_users(search='John')
            >>> mosaic.unshare(users=users)
        """
        super()._unshare(self.endpoint, users)


    def get_shared_users(self, search: str = None, skip: int = 0, limit: int = 10) -> List['User']:
        """
        Retrieves the list of users the file is shared with.

        Args:
            search (str, optional): The search query.
            skip (int, optional): The number of users to skip.
            limit (int, optional): The maximum number of users to retrieve.

        Returns:
            List[User]: The list of shared users.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.mosaic import Mosaic
            >>> client = GeoboxClient()
            >>> mosaic = Mosaic.get_mosaic(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> mosaic.get_shared_users(search='John', skip=0, limit=10)
        """
        params = {
            'search': search,
            'skip': skip,
            'limit': limit
        }
        return super()._get_shared_users(self.endpoint, params)


    def seed_cache(self, 
                   from_zoom: int = None, 
                   to_zoom: int = None, 
                   extent: List[int] = None, 
                   workers: int = 1) -> List['Task']:
        """
        Seed the cache of the mosaic.

        Args:
            from_zoom (int, optional): The from zoom of the mosaic.
            to_zoom (int, optional): The to zoom of the mosaic.
            extent (list[int], optional): The extent of the mosaic.
            workers (int, optional): The number of workers to use. default is 1.

        Returns:
            List[Task]: The task of the seed cache.

        Example:
            >>> from geobox import GeoboxClient 
            >>> from geobox.mosaic import Mosaic
            >>> client = GeoboxClient()
            >>> mosaic = Mosaic.get_mosaic(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> task = mosaic.seed_cache(from_zoom=0, to_zoom=22, extent=[0, 0, 100, 100], workers=1)
        """
        data = {
            'from_zoom': from_zoom,
            'to_zoom': to_zoom,
            'extent': extent,
            'workers': workers
        }
        return super()._seed_cache(endpoint=self.endpoint, data=data)


    def clear_cache(self) -> None:
        """
        Clear the cache of the mosaic.

        Returns:
            None

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.mosaic import Mosaic
            >>> client = GeoboxClient()
            >>> mosaic = Mosaic.get_mosaic(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> mosaic.clear_cache()
        """
        return super().clear_cache()


    @property
    def cache_size(self) -> int:
        """
        Get the size of the cache of the mosaic.

        Returns:
            int: The size of the cache of the mosaic.

        Example:
            >>> from geobox import GeoboxClient
            >>> from geobox.mosaic import Mosaic
            >>> client = GeoboxClient()
            >>> mosaic = Mosaic.get_mosaic(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> mosaic.cache_size
        """
        return super().cache_size
    

    def to_async(self, async_client: 'AsyncGeoboxClient') -> 'AsyncMosaic':
        """
        Switch to async version of the mosaic instance to have access to the async methods

        Args:
            async_client (AsyncGeoboxClient): The async version of the GeoboxClient instance for making requests.

        Returns:
            AsyncMosaic: the async instance of the mosaic.

        Example:
            >>> from geobox import Geoboxclient
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.mosaic import Mosaic
            >>> client = GeoboxClient()
            >>> mosaic = Mosaic.get_mosaic(client, uuid="12345678-1234-5678-1234-567812345678")
            >>> async with AsyncGeoboxClient() as async_client:
            >>>     async_mosaic = mosaic.to_async(async_client)
        """
        from .aio.mosaic import AsyncMosaic

        return AsyncMosaic(api=async_client, uuid=self.uuid, data=self.data)



import pytest
from unittest.mock import patch

from geobox.scene import Scene
from geobox.user import User

def test_init(api, mock_scene_data):
    scene = Scene(api, mock_scene_data['uuid'], mock_scene_data)
    assert scene.name == mock_scene_data['name']
    assert scene.uuid == mock_scene_data['uuid']
    assert scene.data == mock_scene_data
    assert scene.endpoint == f'{Scene.BASE_ENDPOINT}{scene.uuid}/'


def test_get_scenes(api, mock_scene_data):
    api.get.return_value = [mock_scene_data, mock_scene_data]
    workflows = Scene.get_scenes(api)
    api.get.assert_called_once_with(f'{Scene.BASE_ENDPOINT}?f=json&return_count=False&skip=0&limit=10&shared=False')
    assert len(workflows) == 2
    assert type(workflows[0]) == Scene
    assert workflows[0].data == mock_scene_data


def test_create_scene(api, mock_scene_data):
    api.post.return_value = mock_scene_data
    scene = Scene.create_scene(api,
                                name=mock_scene_data['name'],
                                display_name=mock_scene_data['display_name'],
                                description=mock_scene_data['description'],
                                settings=mock_scene_data['settings'],
                                thumbnail=mock_scene_data.get('thumbnail'))
    api.post.assert_called_once_with(Scene.BASE_ENDPOINT, {'name': 'test'})
    assert type(scene) == Scene
    assert scene.uuid == mock_scene_data['uuid']
    assert scene.data == mock_scene_data


def test_get_scene(api, mock_scene_data):
    api.get.return_value = mock_scene_data
    scene = Scene.get_scene(api, uuid=mock_scene_data['uuid'])
    api.get.assert_called_once_with(f"{Scene.BASE_ENDPOINT}{mock_scene_data['uuid']}/?f=json")
    assert type(scene) == Scene
    assert scene.uuid == mock_scene_data['uuid']
    assert scene.data == mock_scene_data


def test_get_scene_by_name(api, mock_scene_data):
    api.get.return_value = [mock_scene_data]
    scene = Scene.get_scene_by_name(api, name=mock_scene_data['name'])
    api.get.assert_called_once_with(f"{Scene.BASE_ENDPOINT}?f=json&q=name+%3D+%27test%27&return_count=False&skip=0&limit=10&shared=False")
    assert type(scene) == Scene
    assert scene.uuid == mock_scene_data['uuid']
    assert scene.data == mock_scene_data
    # not found
    scene = Scene.get_scene_by_name(api, name='not_found')
    assert scene == None


def test_update(api, mock_scene_data):
    scene = Scene(api, mock_scene_data['uuid'], mock_scene_data)
    updated_data = {**mock_scene_data,
                    'settings': {'settings': 'value'},
                    'thumbnail': 'thumbnail-url'}

    api.put.return_value = updated_data
    scene.update(settings={'settings': 'value'}, thumbnail='thumbnail-url')
    api.put.assert_called_once_with(scene.endpoint, {'settings': {'settings': 'value'}, 'thumbnail': 'thumbnail-url'})


def test_delete(api, mock_scene_data):
    scene = Scene(api, mock_scene_data['uuid'], mock_scene_data)
    endpoint = scene.endpoint
    scene.delete()
    api.delete.assert_called_once_with(endpoint)
    assert scene.uuid is None
    assert scene.endpoint is None


def test_thumbnail(api, mock_scene_data):
    scene = Scene(api, mock_scene_data['uuid'], mock_scene_data)
    thumbnail_url = scene.thumbnail
    assert thumbnail_url == f"{api.base_url}{scene.endpoint}thumbnail.png"


def test_share(api, mock_scene_data, mock_user_data):
    """Test file sharing."""
    with patch.object(api, 'post'):
        scene = Scene(api, uuid=mock_scene_data['uuid'], data=mock_scene_data)
        users = [
            User(api, user_id=mock_user_data['1']['id'], data=mock_user_data['1']),
            User(api, user_id=mock_user_data['2']['id'], data=mock_user_data['2'])
        ]
        scene.share(users=users)
        api.post.assert_called_once_with(
            f'{scene.endpoint}share/',
            {'user_ids': [1, 2]}, is_json=False
        )


def test_unshare(api, mock_scene_data, mock_user_data):
    """Test file unsharing."""
    with patch.object(api, 'post'):
        scene = Scene(api, uuid=mock_scene_data['uuid'], data=mock_scene_data)
        users = [
            User(api, user_id=mock_user_data['1']['id'], data=mock_user_data['1']),
            User(api, user_id=mock_user_data['2']['id'], data=mock_user_data['2'])
        ]
        scene.unshare(users=users)
        
        # Verify the API call
        api.post.assert_called_once_with(
            f'{scene.endpoint}unshare/',
            {'user_ids': [1, 2]}, is_json=False
        )


def test_get_shared_users(api, mock_scene_data, mock_user_data):
    """Test getting shared users."""
    mock_response = [
        mock_user_data['1'],
        mock_user_data['2']
    ]
    
    with patch.object(api, 'get', return_value=mock_response):
        scene = Scene(api, uuid=mock_scene_data['uuid'], data=mock_scene_data)
        result = scene.get_shared_users(search='user', limit=2)

        api.get.assert_called_once_with(
            f'{scene.endpoint}shared-with-users/?search=user&skip=0&limit=2'
        )

        assert len(result) == 2
        assert result[0].first_name == 'test 1'
        assert result[0].last_name == 'test 1'
        assert result[1].first_name == 'test 2'
        assert result[1].last_name == 'test 2'


def test_to_async(api, async_api, mock_scene_data):
    scene = Scene(api, uuid=mock_scene_data['uuid'], data=mock_scene_data)
    async_instance = scene.to_async(async_api)
    assert async_instance.api == async_api


class GeoboxError(Exception):
    """Base class for all exceptions raised by the Geobox SDK."""
    pass

# Error Classes
class AuthenticationError(GeoboxError):
    """Raised when there is an authentication error."""

    def __init__(self, message="Authentication failed"):
        self.message = message
        super().__init__(self.message)

class AuthorizationError(GeoboxError):
    """Raised when there is an authorization error."""

    def __init__(self, message="Authorization failed"):
        self.message = message
        super().__init__(self.message)

class ApiRequestError(GeoboxError):
    """Raised when there is an error with the API request."""

    def __init__(self, status_code, message="API request failed"):
        self.status_code = status_code
        self.message = f"{message}: Status code {status_code}"
        super().__init__(self.message)

class NotFoundError(GeoboxError):
    """Raised when a requested resource is not found."""

    def __init__(self, message="Resource not found"):
        self.message = message
        super().__init__(self.message)

class ValidationError(GeoboxError):
    """Raised when there is a validation error."""

    def __init__(self, message="Validation error"):
        self.message = message
        super().__init__(self.message)

class ServerError(GeoboxError):
    """Raised when there is a server error."""
    
    def __init__(self, message="Server error"):
        self.message = message
        super().__init__(self.message)



from urllib.parse import urljoin
from typing import Dict, List, Optional, Union, TYPE_CHECKING

from .base import AsyncBase
from .vectorlayer import AsyncVectorLayer
from .view import AsyncVectorLayerView
from .task import AsyncTask

if TYPE_CHECKING:
    from . import AsyncGeoboxClient 
    from .user import AsyncUser 
    from ..api import GeoboxClient
    from ..tileset import Tileset

class AsyncTileset(AsyncBase):

    BASE_ENDPOINT: str = 'tilesets/'

    def __init__(self, 
        api: 'AsyncGeoboxClient', 
        uuid: str, 
        data: Optional[Dict] = {}):
        """
        Constructs all the necessary attributes for the Tilesets object.

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            uuid (str): The UUID of the tileset.
            data (Dict, optional): The data of the tileset.
        """
        super().__init__(api=api, uuid=uuid, data=data)


    @classmethod
    async def create_tileset(cls, 
        api: 'AsyncGeoboxClient', 
        name: str, 
        layers: List[Union['AsyncVectorLayer', 'AsyncVectorLayerView']], 
        display_name: str = None, 
        description: str = None,
        min_zoom: int = None, 
        max_zoom: int = None, 
        user_id: int = None) -> 'AsyncTileset':
        """
        [async] Create a new tileset.

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            name (str): The name of the tileset.
            layers (List[VectorLayer | VectorLayerView]): list of vectorlayer and view objects to add to tileset.
            display_name (str, optional): The display name of the tileset.
            description (str, optional): The description of the tileset.
            min_zoom (int, optional): The minimum zoom level of the tileset.
            max_zoom (int, optional): The maximum zoom level of the tileset.
            user_id (int, optional): Specific user. privileges required.

        Returns:
            AsyncTileset: The created tileset instance.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.tileset import AsyncTileset
            >>> async with AsyncGeoboxClient() as client:
            >>>     layer = await client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>>     view = await client.get_view(uuid="12345678-1234-5678-1234-567812345678")
            >>>     tileset = await AsyncTileset.create_tileset(client, 
            ...                                             name="your_tileset_name", 
            ...                                             display_name="Your Tileset", 
            ...                                             description="Your description", 
            ...                                             min_zoom=0, 
            ...                                             max_zoom=14, 
            ...                                             layers=[layer, view])
            or  
            >>>     tileset = await client.create_tileset(name="your_tileset_name", 
            ...                                             display_name="Your Tileset", 
            ...                                             description="Your description", 
            ...                                             min_zoom=0, 
            ...                                             max_zoom=14, 
            ...                                             layers=[layer, view])
        """
        formatted_layers = []
        for item in layers:
            if type(item) == AsyncVectorLayer:
                item_type = 'vector'

            elif type(item) == AsyncVectorLayerView:
                item_type = 'view'

            else:
                continue

            formatted_layers.append({
                'layer_type': item_type,
                'layer_uuid': item.uuid
            })

        data = {
            "name": name,
            "display_name": display_name,
            "description": description,
            "min_zoom": min_zoom,
            "max_zoom": max_zoom,
            "layers": formatted_layers,
            "user_id": user_id
        }
        return await super()._create(api, cls.BASE_ENDPOINT, data, factory_func=lambda api, item: AsyncTileset(api, item['uuid'], item))


    @classmethod
    async def get_tilesets(cls, api: 'AsyncGeoboxClient', **kwargs) -> Union[List['AsyncTileset'], int]:
        """
        [async] Retrieves a list of tilesets.

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
        
        Keyword Args:
            q (str): query filter based on OGC CQL standard. e.g. "field1 LIKE '%GIS%' AND created_at > '2021-01-01'"
            search (str): search term for keyword-based searching among search_fields or all textual fields if search_fields does not have value. NOTE: if q param is defined this param will be ignored.
            search_fields (str): comma separated list of fields for searching.
            order_by (str): comma separated list of fields for sorting results [field1 A|D, field2 A|D, …]. e.g. name A, type D. NOTE: "A" denotes ascending order and "D" denotes descending order.
            return_count (bool): if True, returns the total number of tilesets matching the query. default is False.
            skip (int): number of records to skip. default is 0.
            limit (int): number of records to return. default is 10.
            user_id (int): Specific user. privileges required.
            shared (bool): Whether to return shared tilesets. default is False.

        Returns:
            List[AsyncTileset] | int: A list of Tileset instances or the total number of tilesets

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.tileset import AsyncTileset
            >>> async with AsyncGeoboxClient() as client:
            >>>     tilesets = await AsyncTileset.get_tilesets(client,
            ...         q="name LIKE '%your_tileset_name%'",
            ...         order_by="name A",
            ...         skip=0,
            ...         limit=10,
            ...     )
            or  
            >>>     tilesets = await client.get_tilesets(q="name LIKE '%your_tileset_name%'",
            ...         order_by="name A",
            ...         skip=0,
            ...         limit=10,
            ...     )
        """
        params = {
            'f': 'json',
            'q': kwargs.get('q', None),
            'search': kwargs.get('search', None),
            'search_fields': kwargs.get('search_fields', None),
            'order_by': kwargs.get('order_by', None),
            'return_count': kwargs.get('return_count', False),
            'skip': kwargs.get('skip', 0),
            'limit': kwargs.get('limit', 10),
            'user_id': kwargs.get('user_id', None),
            'shared': kwargs.get('shared', False)
        }
        return await super()._get_list(api=api, endpoint=cls.BASE_ENDPOINT, params=params, factory_func=lambda api, item: AsyncTileset(api, item['uuid'], item))
    

    @classmethod
    async def get_tilesets_by_ids(cls, api: 'AsyncGeoboxClient', ids: List[str], user_id: int = None) -> List['AsyncTileset']:
        """
        [async] Retrieves a list of tilesets by their IDs.

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            ids (List[str]): The list of tileset IDs.
            user_id (int, optional): Specific user. privileges required.

        Returns:
            List[AsyncTileset]: A list of Tileset instances.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.tileset import AsyncTileset
            >>> async with AsyncGeoboxClient() as client:
            >>>     tilesets = await AsyncTileset.get_tilesets_by_ids(client, ids=['123', '456'])
            or  
            >>>     tilesets = await client.get_tilesets_by_ids(ids=['123', '456'])
        """
        params = {
            'ids': ids,
            'user_id': user_id
        }
        return await super()._get_list_by_ids(api, cls.BASE_ENDPOINT, params, factory_func=lambda api, item: AsyncTileset(api, item['uuid'], item))


    @classmethod
    async def get_tileset(cls, api: 'AsyncGeoboxClient', uuid: str) -> 'AsyncTileset':
        """
        [async] Retrieves a tileset by its UUID.

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            uuid (str): The UUID of the tileset.

        Returns:
            AsyncTileset: The retrieved tileset instance.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.tileset import AsyncTileset
            >>> async with AsyncGeoboxClient() as client:
            >>>     tileset = await AsyncTileset.get_tileset(client, uuid="12345678-1234-5678-1234-567812345678")
            or  
            >>>     tileset = await client.get_tileset(uuid="12345678-1234-5678-1234-567812345678")
        """
        return await super()._get_detail(api, cls.BASE_ENDPOINT, uuid, factory_func=lambda api, item: AsyncTileset(api, item['uuid'], item))


    @classmethod
    async def get_tileset_by_name(cls, api: 'AsyncGeoboxClient', name: str, user_id: int = None) -> Union['AsyncTileset', None]:
        """
        [async] Get a tileset by name

        Args:
            api (AsyncGeoboxClient): The AsyncGeoboxClient instance for making requests.
            name (str): the name of the tileset to get
            user_id (int, optional): specific user. privileges required.

        Returns:
            AsyncTileset | None: returns the tileset if a tileset matches the given name, else None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.tileset import AsyncTileset
            >>> async with AsyncGeoboxClient() as client:
            >>>     tileset = await VectorLayer.get_tileset_by_name(client, name='test')
            or  
            >>>     tileset = await client.get_tileset_by_name(name='test')
        """
        tilesets = await cls.get_tilesets(api, q=f"name = '{name}'", user_id=user_id)
        if tilesets and tilesets[0].name == name:
            return tilesets[0]
        else:
            return None


    async def update(self, **kwargs) -> None:
        """
        [async] Updates the properties of the tileset.

        Keyword Args:
            name (str): The new name of the tileset.
            display_name (str): The new display name of the tileset.
            description (str): The new description of the tileset.
            min_zoom (int): The new minimum zoom level of the tileset.
            max_zoom (int): The new maximum zoom level of the tileset.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.tileset import AsyncTileset
            >>> async with AsyncGeoboxClient() as client:
            >>>     tileset = await AsyncTileset.get_tileset(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await tileset.update_tileset(
            ...         name="new_name",
            ...         display_name="New Display Name",
            ...         description="New description",
            ...         min_zoom=0,
            ...         max_zoom=14
            ...     )
        """
        data = {
            "name": kwargs.get('name'),
            "display_name": kwargs.get('display_name'),
            "description": kwargs.get('description'),
            "min_zoom": kwargs.get('min_zoom'),
            "max_zoom": kwargs.get('max_zoom')
        }

        return await super()._update(urljoin(self.BASE_ENDPOINT, f'{self.uuid}/'), data)


    async def delete(self) -> None:
        """
        [async] Deletes the tileset.

        Raises:
            ValueError: if the tileset uuid is not set.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.tileset import AsyncTileset
            >>> async with AsyncGeoboxClient() as client:
            >>>     tileset = await AsyncTileset.get_tileset(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await tileset.delete()
        """
        await super()._delete(urljoin(self.BASE_ENDPOINT, f'{self.uuid}/'))
    
    
    async def get_layers(self, **kwargs) -> Union[List['AsyncVectorLayer'], List['AsyncVectorLayerView']]:
        """
        [async] Retrieves the layers of the tileset with optional parameters.

        Keyword Args:
            q (str): query filter based on OGC CQL standard. e.g. "field1 LIKE '%GIS%' AND created_at > '2021-01-01'"
            search (str): search term for keyword-based searching among search_fields or all textual fields if search_fields does not have value. NOTE: if q param is defined this param will be ignored.
            search_fields (str): comma separated list of fields for searching.
            order_by (str): comma separated list of fields for sorting results [field1 A|D, field2 A|D, …]. e.g. name A, type D. NOTE: "A" denotes ascending order and "D" denotes descending order.
            return_count (bool): if True, returns the total number of layers matching the query. default is False.
            skip (int): number of records to skip. default is 0.
            limit (int): number of records to return. default is 10.
            user_id (int): specific user. privileges required.
            shared (bool): if True, returns only the layers that has been shared with you. default is False.

        Returns:
            AsyncVectorLayer | AsyncVectorLayerView: A list of VectorLayer or VectorLayerView instances.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.tileset import AsyncTileset
            >>> async with AsyncGeoboxClient() as client:
            >>>     tileset = await AsyncTileset.get_tileset(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     layers = await tileset.get_layers()
        """
        params = {
            'f': 'json',
            'q': kwargs.get('q'),
            'search': kwargs.get('search'),
            'seacrh_fields': kwargs.get('seacrh_fields'),
            'order_by': kwargs.get('order_by'),
            'return_count': kwargs.get('return_count', False),
            'skip': kwargs.get('skip', 0),
            'limit': kwargs.get('limit', 10),
            'user_id': kwargs.get('user_id'),
            'shared': kwargs.get('shared', False)
        }
        endpoint = urljoin(self.BASE_ENDPOINT, f'{self.uuid}/layers/')
        return await super()._get_list(self.api, endpoint, params, factory_func=lambda api, item: AsyncVectorLayer(api, item['uuid'], item['layer_type'], item) if not item['is_view'] else \
                                                                                            AsyncVectorLayerView(api, item['uuid'], item['layer_type'], item))


    async def add_layer(self, layer: Union['AsyncVectorLayer', 'AsyncVectorLayerView']) -> None:
        """
        [async] Adds a layer to the tileset.

        Args:
            layer (AsyncVectorLayer | AsyncVectorLayerView): the layer object to add to the tileset

        Returns:
            None

        Raises:
            ValueError: if the layer input is not a 'VectorLayer' or 'VetorLayerview' object

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.tileset import AsyncTileset
            >>> async with AsyncGeoboxClient() as client:
            >>>     tileset = await AsyncTileset.get_tileset(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     layer = await client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>>     await tileset.add_layer(layer)
        """
        if type(layer) == AsyncVectorLayer:
            layer_type = 'vector'

        elif type(layer) == AsyncVectorLayerView:
            layer_type = 'view'

        else:
            raise ValueError("layer input must be either 'VectorLayer' or 'VetorLayerview' object")

        data = {
            "layer_type": layer_type,
            "layer_uuid": layer.uuid
        }

        endpoint = urljoin(self.endpoint, 'layers/')
        return await self.api.post(endpoint, data, is_json=False)


    async def delete_layer(self, layer: Union['AsyncVectorLayer', 'AsyncVectorLayerView']) -> None:
        """
        [async] Deletes a layer from the tileset.

        Args:
            layer (AsyncVectorLayer | AsyncVectorLayerView): the layer object to delete from the tileset

        Returns:
            None

        Raises:
            ValueError: if the layer input is not a 'VectorLayer' or 'VetorLayerview' object

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.tileset import AsyncTileset
            >>> async with AsyncGeoboxClient() as client:
            >>>     tileset = await AsyncTileset.get_tileset(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     layer = await client.get_vector(uuid="12345678-1234-5678-1234-567812345678")
            >>>     await tileset.delete_layer(layer)
        """
        if type(layer) == AsyncVectorLayer:
            layer_type = 'vector'

        elif type(layer) == AsyncVectorLayerView:
            layer_type = 'view'

        else:
            raise ValueError("layer input must be either 'VectorLayer' or 'VetorLayerview' object")

        data = {
            "layer_type": layer_type,
            "layer_uuid": layer.uuid
        }

        endpoint = urljoin(self.endpoint, 'layers/')
        return await self.api.delete(endpoint, data, is_json=False)


    async def share(self, users: List['AsyncUser']) -> None:
        """
        [async] Shares the file with specified users.

        Args:
            users (List[AsyncUser]): The list of user IDs to share the file with.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.tileset import AsyncTileset
            >>> async with AsyncGeoboxClient() as client:
            >>>     tileset = await AsyncTileset.get_tileset(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     users = await client.search_usesrs(search='John')
            >>>     await tileset.share(users=users)
        """
        await super()._share(self.endpoint, users)
    

    async def unshare(self, users: List['AsyncUser']) -> None:
        """
        [async] Unshares the file with specified users.

        Args:
            users (List[AsyncUser]): The list of user IDs to unshare the file with.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.tileset import AsyncTileset
            >>> async with AsyncGeoboxClient() as client:
            >>>     tileset = await AsyncTileset.get_tileset(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     users = await client.search_usesrs(search='John')
            >>>     await tileset.unshare(users=users)
        """
        await super()._unshare(self.endpoint, users)


    async def get_shared_users(self, search: str = None, skip: int = 0, limit: int = 10) -> List['AsyncUser']:
        """
        [async] Retrieves the list of users the file is shared with.

        Args:
            search (str, optional): The search query.
            skip (int, optional): The number of users to skip.
            limit (int, optional): The maximum number of users to retrieve.

        Returns:
            List[AsyncUser]: The list of shared users.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.tileset import AsyncTileset
            >>> async with AsyncGeoboxClient() as client:
            >>>     tileset = await AsyncTileset.get_tileset(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await tileset.get_shared_users(search='John', skip=0, limit=10)
        """
        params = {
            'search': search,
            'skip': skip,
            'limit': limit
        }
        return await super()._get_shared_users(self.endpoint, params)
        

    async def get_tile_json(self) -> Dict:
        """
        [async] Retrieves the tile JSON configuration.

        Returns:
            Dict: The tile JSON configuration.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.tileset import AsyncTileset
            >>> async with AsyncGeoboxClient() as client:
            >>>     tileset = await AsyncTileset.get_tileset(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await tileset.get_tile_json()
        """ 
        endpoint = urljoin(self.endpoint, 'tilejson.json/')
        return await self.api.get(endpoint)


    async def update_tileset_extent(self) -> Dict:
        """
        [async] Updates the extent of the tileset.

        Returns:
            Dict: The response from the API.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.tileset import AsyncTileset
            >>> async with AsyncGeoboxClient() as client:
            >>>     tileset = await AsyncTileset.get_tileset(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await tileset.update_tileset_extent()
        """
        endpoint = urljoin(self.endpoint, 'updateExtent/')
        return await self.api.post(endpoint)


    def get_tile_pbf_url(self, x: 'int' = '{x}', y: int = '{y}', z: int = '{z}') -> str:
        """
        Retrieves a tile from the tileset.

        Args:
            x (int, optional): The x coordinate of the tile.
            y (int, optioanl): The y coordinate of the tile.
            z (int, optional): The zoom level of the tile.

        Returns:
            str: The url of the tile.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.tileset import AsyncTileset
            >>> async with AsyncGeoboxClient() as client:
            >>>     tileset = await AsyncTileset.get_tileset(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await tileset.get_tile_tileset(x=1, y=1, z=1)
        """
        endpoint = f'{self.api.base_url}{self.endpoint}tiles/{z}/{x}/{y}.pbf'

        if not self.api.access_token and self.api.apikey:
            endpoint = f'{endpoint}?apikey={self.api.apikey}'

        return endpoint


    async def seed_cache(self, from_zoom: int = 0, to_zoom: int = 14, extent: list = [], workers: int = 1, user_id: int = 0) -> List['AsyncTask']:
        """
        [async] Seeds the cache of the tileset.

        Args:
            from_zoom (int, optional): The starting zoom level.
            to_zoom (int, optional): The ending zoom level.
            extent (list, optional): The extent of the tileset.
            workers (int, optional): The number of workers to use.
            user_id (int, optional): The user ID.

        Returns:
            List[AsyncTask]: list of task objects.

        Raises:
            ValueError: If the number of workers is not one of the following: 1, 2, 4, 8, 12, 16, 20, 24.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.tileset import AsyncTileset
            >>> async with AsyncGeoboxClient() as client:
            >>>     tileset = await AsyncTileset.get_tileset(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await tileset.seed_cache(from_zoom=0, to_zoom=14, extent=[], workers=1)
        """
        data = {
            "from_zoom": from_zoom,
            "to_zoom": to_zoom,
            "extent": extent,
            "workers": workers,
            "user_id": user_id
        }
        return await super()._seed_cache(endpoint=self.endpoint, data=data)


    async def update_cache(self, from_zoom: int, to_zoom: int, extents: List[List[float]] = None, user_id: int = 0) -> List['AsyncTask']:
        """
        [async] Updates the cache of the tileset.

        Args:
            from_zoom (int): The starting zoom level.
            to_zoom (int): The ending zoom level.
            extents (List[List[float]], optional): The list of extents to update the cache for.
            user_id (int, optional): The user ID.

        Returns:
            List[Task]: list of task objects.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.tileset import AsyncTileset
            >>> async with AsyncGeoboxClient() as client:
            >>>     tileset = await AsyncTileset.get_tileset(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await tileset.update_cache(from_zoom=0, to_zoom=14, extent=[])
        """
        data = {
            "from_zoom": from_zoom,
            "to_zoom": to_zoom,
            "extents": extents,
            "user_id": user_id
        }
        return await super()._update_cache(endpoint=self.endpoint, data=data)


    @property
    async def cache_size(self) -> int:
        """
        [async] Retrieves the size of the cache of the tileset.

        Returns:
            int: The size of the cache of the tileset.

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.tileset import AsyncTileset
            >>> async with AsyncGeoboxClient() as client:
            >>>     tileset = await AsyncTileset.get_tileset(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await tileset.cache_size
        """
        return await super()._cache_size(endpoint=self.endpoint)


    async def clear_cache(self) -> None:
        """
        [async] Clears the cache of the tileset.

        Returns:
            None

        Example:
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.tileset import AsyncTileset
            >>> async with AsyncGeoboxClient() as client:
            >>>     tileset = await AsyncTileset.get_tileset(client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     await tileset.clear_cache()
        """
        await super()._clear_cache(endpoint=self.endpoint)


    def to_sync(self, sync_client: 'GeoboxClient') -> 'Tileset':
        """
        Switch to sync version of the tileset instance to have access to the sync methods

        Args:
            sync_client (GeoboxClient): The sync version of the GeoboxClient instance for making requests.

        Returns:
            Tileset: the sync instance of the tileset.

        Example:
            >>> from geobox import Geoboxclient
            >>> from geobox.aio import AsyncGeoboxClient
            >>> from geobox.aio.tileset import AsyncTileset
            >>> client = GeoboxClient()
            >>> async with AsyncGeoboxClient() as async_client:
            >>>     tileset = await AsyncTileset.get_tileset(async_client, uuid="12345678-1234-5678-1234-567812345678")
            >>>     sync_tileset = tileset.to_sync(client)
        """
        from ..tileset import Tileset

        return Tileset(api=sync_client, uuid=self.uuid, data=self.data)